import time
import threading
from Pyro5.api import expose, oneway, behavior, serve


@expose
@behavior("single")
class Server(object):
    def __init__(self):
        self.counter = 0

    @oneway
    def increment_oneway(self):
        print("oneway call executing in thread", threading.get_ident())
        time.sleep(0.5)
        self.counter += 1

    def increment(self):
        time.sleep(0.5)
        self.counter += 1

    def getcount(self):
        return self.counter


print("main thread:", threading.get_ident())
serve({
    Server: "example.oneway2"
})



import datetime
import pytz
import dateutil.tz
import pendulum
from Pyro5.api import expose, serve


fmt = '%Y-%m-%d %H:%M:%S %Z%z'


@expose
class Server(object):
    def echo(self, date):
        print("RETURNING:", repr(date))
        return date

    def pytz(self):
        tz_nl = pytz.timezone("Europe/Amsterdam")
        result = tz_nl.localize(datetime.datetime.now())
        print("RETURNING:", repr(result))
        return result

    def dateutil(self):
        tz_nl = dateutil.tz.gettz("Europe/Amsterdam")
        result = datetime.datetime.now(tz_nl)
        print("RETURNING:", repr(result))
        return result

    def pendulum(self):
        tz_nl = pendulum.now("Europe/Amsterdam")
        print("RETURNING:", repr(tz_nl))
        return tz_nl


# main program

serve({
    Server: "example.timezones"
}, use_ns=False)



import sys
import Pyro5.errors
import Pyro5.api


sys.excepthook = Pyro5.errors.excepthook


uri = input("Enter streaming server uri: ").strip()
with Pyro5.api.Proxy(uri) as p:
    print("\nnormal list:")
    print(p.list())
    print("\nvia iterator:")
    print(list(p.iterator()))
    print("\nvia generator:")
    print(list(p.generator()))
    print("\nslow generator:")
    for number in p.slow_generator():
        print(number)
    print("\nproperty iterator:")
    print(list(p.prop_iter))
    print("\nproperty generator:")
    print(list(p.prop_generator))



# Client that doesn't use the Name Server. Uses URI directly.

from Pyro5.api import Proxy


uri = input("Enter the URI of the quote object: ")
with Proxy(uri) as quotegen:
    print("Getting some quotes...")
    print(quotegen.quote())
    print(quotegen.quote())



# This is the code that visits the warehouse.
from Pyro5.api import Proxy
from person import Person


uri = input("Enter the uri of the warehouse: ").strip()
warehouse = Proxy(uri)
janet = Person("Janet")
henry = Person("Henry")
janet.visit(warehouse)
henry.visit(warehouse)



import sys
import os
import platform
import threading
import socket
import time
import ssl
import pytest
import contextlib
from Pyro5 import config, socketutil, protocol, errors, server, serializers, core
from Pyro5.svr_threads import SocketServer_Threadpool
from Pyro5.svr_multiplex import SocketServer_Multiplex


# determine ipv6 capability
has_ipv6 = socket.has_ipv6
if has_ipv6:
    s = socket.socket(socket.AF_INET6, socket.SOCK_DGRAM)
    try:
        s.connect(("::1", 53))
        s.close()
        socket.getaddrinfo("localhost", 53, socket.AF_INET6)
    except socket.error:
        has_ipv6 = False


class TestSocketutil:
    @classmethod
    def setup_class(cls):
        config.POLLTIMEOUT = 0.1

    @pytest.mark.network
    def testGetIP(self):
        config.PREFER_IP_VERSION = 4
        myip = socketutil.get_ip_address("")
        assert len(str(myip)) > 4
        myip = socketutil.get_ip_address("", workaround127=True)
        assert len(str(myip)) > 4
        assert not str(myip).startswith("127.")
        addr = socketutil.get_ip_address("127.0.0.1", workaround127=False)
        assert "127.0.0.1" == str(addr)
        assert addr.version == 4
        addr = socketutil.get_ip_address("127.0.0.1", workaround127=True)
        assert "127.0.0.1" != str(addr)
        assert addr.version == 4

    def testGetIP6(self):
        if not has_ipv6:
            pytest.skip("no ipv6 capability")
        addr = socketutil.get_ip_address("::1", version=6)
        assert addr.version == 6
        assert ":" in str(addr)
        addr = socketutil.get_ip_address("localhost", version=6)
        assert addr.version == 6
        assert ":" in str(addr)

    def testGetInterface(self):
        addr = socketutil.get_interface("localhost")
        if addr.version == 6:
            assert str(addr).startswith("::1")
            assert str(addr.ip) == "::1"
            assert str(addr.network).startswith("::1")
        else:
            assert addr.version == 4
            assert str(addr).startswith("127.")
            assert str(addr.ip).startswith("127.0")
            assert str(addr.network).startswith("127.0")
        if has_ipv6:
            addr = socketutil.get_interface("::1")
            assert addr.version == 6
            assert ":" in str(addr)
            assert ":" in str(addr.ip)
            assert ":" in str(addr.network)

    def testUnusedPort(self):
        port1 = socketutil.find_probably_unused_port()
        port2 = socketutil.find_probably_unused_port()
        assert port1 > 0
        assert port1 != port2
        port1 = socketutil.find_probably_unused_port(socktype=socket.SOCK_DGRAM)
        port2 = socketutil.find_probably_unused_port(socktype=socket.SOCK_DGRAM)
        assert port1 > 0
        assert port1 != port2

    def testUnusedPort6(self):
        if not has_ipv6:
            pytest.skip("no ipv6 capability")
        port1 = socketutil.find_probably_unused_port(family=socket.AF_INET6)
        port2 = socketutil.find_probably_unused_port(family=socket.AF_INET6)
        assert port1 > 0
        assert port1 != port2
        port1 = socketutil.find_probably_unused_port(family=socket.AF_INET6, socktype=socket.SOCK_DGRAM)
        port2 = socketutil.find_probably_unused_port(family=socket.AF_INET6, socktype=socket.SOCK_DGRAM)
        assert port1 > 0
        assert port1 != port2

    def testBindUnusedPort(self):
        sock1 = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock2 = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        port1 = socketutil.bind_unused_port(sock1)
        port2 = socketutil.bind_unused_port(sock2)
        assert port1 > 0
        assert port1 != port2
        assert sock1.getsockname() == ("127.0.0.1", port1)
        sock1.close()
        sock2.close()

    def testBindUnusedPort6(self):
        if not has_ipv6:
            pytest.skip("no ipv6 capability")
        sock1 = socket.socket(socket.AF_INET6, socket.SOCK_STREAM)
        sock2 = socket.socket(socket.AF_INET6, socket.SOCK_STREAM)
        port1 = socketutil.bind_unused_port(sock1)
        port2 = socketutil.bind_unused_port(sock2)
        assert port1 > 0
        assert port1 != port2
        host, port, _, _ = sock1.getsockname()
        assert ":" in host
        assert port1 == port
        sock1.close()
        sock2.close()

    def testCreateUnboundSockets(self):
        s = socketutil.create_socket()
        assert socket.AF_INET == s.family
        bs = socketutil.create_bc_socket()
        assert socket.AF_INET == bs.family
        with contextlib.suppress(socket.error):
            host, port = s.getsockname()
            # can either fail with socket.error or return (host,0)
            assert 0 == port
        with contextlib.suppress(socket.error):
            host, port = bs.getsockname()
            # can either fail with socket.error or return (host,0)
            assert 0 == port
        s.close()
        bs.close()

    def testCreateUnboundSockets6(self):
        if not has_ipv6:
            pytest.skip("no ipv6 capability")
        s = socketutil.create_socket(ipv6=True)
        assert socket.AF_INET6 == s.family
        bs = socketutil.create_bc_socket(ipv6=True)
        assert socket.AF_INET6 == bs.family
        with contextlib.suppress(socket.error):
            host, port, _, _ = s.getsockname()
            # can either fail with socket.error or return (host,0)
            assert 0 == port
        with contextlib.suppress(socket.error):
            host, port, _, _ = bs.getsockname()
            # can either fail with socket.error or return (host,0)
            assert 0 == port
        s.close()
        bs.close()

    def testCreateBoundSockets(self):
        s = socketutil.create_socket(bind=('127.0.0.1', 0))
        assert socket.AF_INET == s.family
        bs = socketutil.create_bc_socket(bind=('127.0.0.1', 0))
        assert '127.0.0.1' == s.getsockname()[0]
        assert '127.0.0.1' == bs.getsockname()[0]
        s.close()
        bs.close()
        with pytest.raises(ValueError):
            socketutil.create_socket(bind=('localhost', 12345), connect=('localhost', 1234))

    def testCreateBoundSockets6(self):
        if not has_ipv6:
            pytest.skip("no ipv6 capability")
        s = socketutil.create_socket(bind=('::1', 0))
        assert socket.AF_INET6 == s.family
        bs = socketutil.create_bc_socket(bind=('::1', 0))
        assert ':' in s.getsockname()[0]
        assert ':' in bs.getsockname()[0]
        s.close()
        bs.close()
        with pytest.raises(ValueError):
            socketutil.create_socket(bind=('::1', 12345), connect=('::1', 1234))

    def testCreateBoundUnixSockets(self):
        if not hasattr(socket, "AF_UNIX"):
            pytest.skip("no unix domain sockets capability")
        SOCKNAME = "test_unixsocket"
        if os.path.exists(SOCKNAME):
            os.remove(SOCKNAME)
        s = socketutil.create_socket(bind=SOCKNAME)
        assert socket.AF_UNIX == s.family
        assert SOCKNAME == s.getsockname()
        s.close()
        if os.path.exists(SOCKNAME):
            os.remove(SOCKNAME)
        with pytest.raises(ValueError):
            socketutil.create_socket(bind=SOCKNAME, connect=SOCKNAME)

    def testAbstractNamespace(self):
        if not hasattr(socket, "AF_UNIX") and not sys.platform.startswith("linux"):
            pytest.skip("no unix domain sockets capability, and not Linux")
        SOCKNAME = "\0test_unixsocket_abstract_ns"  # mind the \0 at the start
        s = socketutil.create_socket(bind=SOCKNAME)
        assert bytes(SOCKNAME, "ascii") == s.getsockname()
        s.close()

    def testSend(self):
        ss = socketutil.create_socket(bind=("localhost", 0))
        port = ss.getsockname()[1]
        cs = socketutil.create_socket(connect=("localhost", port))
        socketutil.send_data(cs, b"foobar!" * 10)
        cs.shutdown(socket.SHUT_WR)
        a = ss.accept()
        data = socketutil.receive_data(a[0], 5)
        assert b"fooba" == data
        data = socketutil.receive_data(a[0], 5)
        assert b"r!foo" == data
        a[0].close()
        ss.close()
        cs.close()

    def testSendUnix(self):
        if not hasattr(socket, "AF_UNIX"):
            pytest.skip("no unix domain sockets capability")
        SOCKNAME = "test_unixsocket"
        if os.path.exists(SOCKNAME):
            os.remove(SOCKNAME)
        ss = socketutil.create_socket(bind=SOCKNAME)
        cs = socketutil.create_socket(connect=SOCKNAME)
        socketutil.send_data(cs, b"foobar!" * 10)
        cs.shutdown(socket.SHUT_WR)
        a = ss.accept()
        data = socketutil.receive_data(a[0], 5)
        assert b"fooba" == data
        data = socketutil.receive_data(a[0], 5)
        assert b"r!foo" == data
        a[0].close()
        ss.close()
        cs.close()
        if os.path.exists(SOCKNAME):
            os.remove(SOCKNAME)

    @pytest.mark.network
    def testBroadcast(self):
        ss = socketutil.create_bc_socket((None, 0))
        port = ss.getsockname()[1]
        cs = socketutil.create_bc_socket()
        for bcaddr in config.BROADCAST_ADDRS:
            try:
                cs.sendto(b"monkey", 0, (bcaddr, port))
            except socket.error as x:
                err = getattr(x, "errno", x.args[0])
                # handle some errno that some platforms like to throw
                if err not in socketutil.ERRNO_EADDRNOTAVAIL and err not in socketutil.ERRNO_EADDRINUSE:
                    raise
        data, _ = ss.recvfrom(500)
        assert b"monkey" == data
        cs.close()
        ss.close()

    def testMsgWaitallProblems(self):
        ss = socketutil.create_socket(bind=("localhost", 0), timeout=2)
        port = ss.getsockname()[1]
        cs = socketutil.create_socket(connect=("localhost", port), timeout=2)
        a = ss.accept()
        # test some sizes that might be problematic with MSG_WAITALL and check that they work fine
        for size in [1000, 10000, 32000, 32768, 32780, 41950, 41952, 42000, 65000, 65535, 65600, 80000]:
            socketutil.send_data(cs, b"x" * size)
            data = socketutil.receive_data(a[0], size)
            socketutil.send_data(a[0], data)
            data = socketutil.receive_data(cs, size)
            assert size == len(data)
        a[0].close()
        ss.close()
        cs.close()

    def testMsgWaitallProblems2(self):
        class ReceiveThread(threading.Thread):
            def __init__(self, sock, sizes):
                super(ReceiveThread, self).__init__()
                self.sock = sock
                self.sizes = sizes

            def run(self):
                cs, _ = self.sock.accept()
                for size in self.sizes:
                    data = socketutil.receive_data(cs, size)
                    socketutil.send_data(cs, data)
                cs.close()

        ss = socketutil.create_socket(bind=("localhost", 0))
        SIZES = [1000, 10000, 32000, 32768, 32780, 41950, 41952, 42000, 65000, 65535, 65600, 80000, 999999]
        serverthread = ReceiveThread(ss, SIZES)
        serverthread.daemon = True
        serverthread.start()
        port = ss.getsockname()[1]
        cs = socketutil.create_socket(connect=("localhost", port), timeout=2)
        # test some sizes that might be problematic with MSG_WAITALL and check that they work fine
        for size in SIZES:
            socketutil.send_data(cs, b"x" * size)
            data = socketutil.receive_data(cs, size)
            assert size == len(data)
        serverthread.join()
        ss.close()
        cs.close()

    def testMsgWaitAllConfig(self):
        if platform.system() == "Windows":
            # default config should be False on these platforms even though socket.MSG_WAITALL might exist
            assert not socketutil.USE_MSG_WAITALL
        else:
            # on all other platforms, default config should be True (as long as socket.MSG_WAITALL exists)
            if hasattr(socket, "MSG_WAITALL"):
                assert socketutil.USE_MSG_WAITALL
            else:
                assert not socketutil.USE_MSG_WAITALL



class ServerTestDaemon(server.Daemon):
    pass


class ServerCallback(object):
    def _handshake(self, connection, denied_reason=None):
        raise RuntimeError("this handshake method should never be called")

    def handleRequest(self, connection):
        if not isinstance(connection, socketutil.SocketConnection):
            raise TypeError("handleRequest expected SocketConnection parameter")
        msg = protocol.recv_stub(connection, [protocol.MSG_PING])
        if msg.type == protocol.MSG_PING:
            msg = protocol.SendingMessage(protocol.MSG_PING, 0, msg.seq, msg.serializer_id, b"ping")
            connection.send(msg.data)
        else:
            print("unhandled message type", msg.type)
            connection.close()

    def _housekeeping(self):
        pass


class ServerCallback_BrokenHandshake(ServerCallback):
    def _handshake(self, connection, denied_reason=None):
        raise ZeroDivisionError("handshake crashed (on purpose)")


class TestSocketServer:
    def testServer_thread(self):
        daemon = ServerCallback()
        port = socketutil.find_probably_unused_port()
        serv = SocketServer_Threadpool()
        serv.init(daemon, "localhost", port)
        assert serv.locationStr == "localhost:" + str(port)
        assert serv.sock is not None
        conn = socketutil.SocketConnection(serv.sock, "ID12345")
        assert conn.objectId == "ID12345"
        assert conn.sock is not None
        conn.close()
        conn.close()
        assert conn.sock is not None, "connections keep their socket object even if it's closed"
        serv.close()
        serv.close()
        assert serv.sock is None

    def testServer_multiplex(self):
        daemon = ServerCallback()
        port = socketutil.find_probably_unused_port()
        serv = SocketServer_Multiplex()
        serv.init(daemon, "localhost", port)
        assert serv.locationStr == "localhost:" + str(port)
        assert serv.sock is not None
        conn = socketutil.SocketConnection(serv.sock, "ID12345")
        assert conn.objectId == "ID12345"
        assert conn.sock is not None
        conn.close()
        conn.close()
        assert conn.sock is not None, "connections keep their socket object even if it's closed"
        serv.close()
        serv.close()
        assert serv.sock is None


class TestServerDOS_multiplex:
    def setup_method(self):
        self.orig_poll_timeout = config.POLLTIMEOUT
        self.orig_comm_timeout = config.COMMTIMEOUT
        config.POLLTIMEOUT = 0.5
        config.COMMTIMEOUT = 0.5
        self.socket_server = SocketServer_Multiplex

    def teardown_method(self):
        config.POLLTIMEOUT = self.orig_poll_timeout
        config.COMMTIMEOUT = self.orig_comm_timeout

    class ServerThread(threading.Thread):
        def __init__(self, server, daemon):
            threading.Thread.__init__(self)
            self.serv = server()
            self.serv.init(daemon(), "localhost", 0)
            self.locationStr = self.serv.locationStr
            self.stop_loop = threading.Event()

        def run(self):
            self.serv.loop(loopCondition=lambda: not self.stop_loop.is_set())
            self.serv.close()

    def testConnectCrash(self):
        serv_thread = TestServerDOS_multiplex.ServerThread(self.socket_server, ServerCallback_BrokenHandshake)
        serv_thread.start()
        time.sleep(0.2)
        assert serv_thread.is_alive(), "server thread failed to start"
        threadpool = getattr(serv_thread.serv, "pool", None)
        if threadpool:
            assert len(threadpool.idle) == 1
            assert len(threadpool.busy) == 0
        try:
            host, port = serv_thread.locationStr.split(':')
            port = int(port)
            try:
                # first connection attempt (will fail because server daemon _handshake crashes)
                csock = socketutil.create_socket(connect=(host, port))
                conn = socketutil.SocketConnection(csock, "uri")
                protocol.recv_stub(conn, [protocol.MSG_CONNECTOK])
            except errors.ConnectionClosedError:
                pass
            conn.close()
            time.sleep(0.1)
            if threadpool:
                assert len(threadpool.idle) == 1
                assert len(threadpool.busy) == 0
            try:
                # second connection attempt, should still work (i.e. server should still be running)
                csock = socketutil.create_socket(connect=(host, port))
                conn = socketutil.SocketConnection(csock, "uri")
                protocol.recv_stub(conn, [protocol.MSG_CONNECTOK])
            except errors.ConnectionClosedError:
                pass
        finally:
            if conn:
                conn.close()
            serv_thread.stop_loop.set()
            serv_thread.join()

    def testInvalidMessageCrash(self):
        serv_thread = TestServerDOS_multiplex.ServerThread(self.socket_server, ServerTestDaemon)
        serv_thread.start()
        time.sleep(0.2)
        assert serv_thread.is_alive(), "server thread failed to start"
        threadpool = getattr(serv_thread.serv, "pool", None)
        if threadpool:
            assert len(threadpool.idle) == 1
            assert len(threadpool.busy) == 0

        def connect(host, port):
            # connect to the server
            csock = socketutil.create_socket(connect=(host, port))
            conn = socketutil.SocketConnection(csock, "uri")
            # send the handshake/connect data
            ser = serializers.serializers_by_id[serializers.MarshalSerializer.serializer_id]
            data = ser.dumps({"handshake": "hello", "object": core.DAEMON_NAME})
            msg = protocol.SendingMessage(protocol.MSG_CONNECT, 0, 0, serializers.MarshalSerializer.serializer_id, data)
            conn.send(msg.data)
            # get the handshake/connect response
            protocol.recv_stub(conn, [protocol.MSG_CONNECTOK])
            return conn

        conn = None
        try:
            host, port = serv_thread.locationStr.split(':')
            port = int(port)
            conn = connect(host, port)
            # invoke something, but screw up the message (in this case, mess with the protocol version)
            orig_protocol_version = protocol.PROTOCOL_VERSION
            protocol.PROTOCOL_VERSION = 9999
            msgbytes = protocol.SendingMessage(protocol.MSG_PING, 42, 0, 0, b"something").data
            protocol.PROTOCOL_VERSION = orig_protocol_version
            conn.send(msgbytes)  # this should cause an error in the server because of invalid msg
            try:
                msg = protocol.recv_stub(conn, [protocol.MSG_RESULT])
                data = msg.data.decode("ascii", errors="ignore")  # convert raw message to string to check some stuff
                assert "Traceback" in data
                assert "ProtocolError" in data
                assert "version" in data
            except errors.ConnectionClosedError:
                # invalid message can cause the connection to be closed, this is fine
                pass
            # invoke something again, this should still work (server must still be running, but our client connection was terminated)
            conn.close()
            time.sleep(0.1)
            if threadpool:
                assert len(threadpool.idle) == 1
                assert len(threadpool.busy) == 0
            conn = connect(host, port)
            msg = protocol.SendingMessage(protocol.MSG_PING, 42, 999, 0, b"something")  # a valid message this time
            conn.send(msg.data)
            msg = protocol.recv_stub(conn, [protocol.MSG_PING])
            assert msg.type == protocol.MSG_PING
            assert msg.seq == 999
            assert msg.data == b"pong"
        finally:
            if conn:
                conn.close()
            serv_thread.stop_loop.set()
            serv_thread.join()


class TestServerDOS_threading(TestServerDOS_multiplex):
    def setup_method(self):
        super().setup_method()
        self.socket_server = SocketServer_Threadpool
        self.orig_numthreads = config.THREADPOOL_SIZE
        self.orig_numthreads_min = config.THREADPOOL_SIZE_MIN
        config.THREADPOOL_SIZE = 1
        config.THREADPOOL_SIZE_MIN = 1

    def teardown_method(self):
        config.THREADPOOL_SIZE = self.orig_numthreads
        config.THREADPOOL_SIZE_MIN = self.orig_numthreads_min


class TestSSL:
    def testContextAndSock(self):
        cert_dir = "../../certs"
        if not os.path.isdir(cert_dir):
            cert_dir = "../certs"
            if not os.path.isdir(cert_dir):
                cert_dir = "./certs"
                if not os.path.isdir(cert_dir):
                    raise IOError("cannot locate test certs directory")
        try:
            config.SSL = True
            config.SSL_REQUIRECLIENTCERT = True
            server_ctx = socketutil.get_ssl_context(cert_dir+"/server_cert.pem", cert_dir+"/server_key.pem")
            client_ctx = socketutil.get_ssl_context(clientcert=cert_dir+"/client_cert.pem", clientkey=cert_dir+"/client_key.pem")
            assert server_ctx.verify_mode == ssl.CERT_REQUIRED
            assert client_ctx.verify_mode == ssl.CERT_REQUIRED
            assert client_ctx.check_hostname
            sock = socketutil.create_socket(sslContext=server_ctx)
            try:
                assert hasattr(sock, "getpeercert")
            finally:
                sock.close()
        finally:
            config.SSL = False



# ascii animation of zooming a mandelbrot fractal,  z=z^2+c
import os
import time
import platform
from concurrent import futures
from Pyro5.api import locate_ns, Proxy, BatchProxy


class MandelZoomer(object):
    res_x = 100
    res_y = 40

    def __init__(self):
        self.result = []
        with locate_ns() as ns:
            mandels = ns.yplookup(meta_any={"class:mandelbrot_calc"})
            self.mandels = [uri for _, (uri, meta) in mandels.items()]
        print("{0} mandelbrot calculation servers found.".format(len(self.mandels)))
        if not mandels:
            raise ValueError("launch at least one mandelbrot calculation server before starting this")
        time.sleep(2)

    def screen(self, start, width):
        dr = width / self.res_x
        di = dr*(self.res_x/self.res_y)
        di *= 0.8   # aspect ratio correction
        self.result = ["?"] * self.res_y
        servers = [BatchProxy(Proxy(uri)) for uri in self.mandels]
        with futures.ThreadPoolExecutor(max_workers=len(servers)*2) as pool:
            for i in range(self.res_y):
                server = servers[i % len(servers)]
                server.calc_line(start, self.res_x, i*di, dr, i)
            tasks = [pool.submit(server) for server in servers]
            for task in futures.as_completed(tasks):
                lines = task.result()
                for (linenr, line) in lines:
                    self.result[linenr] = line
        return "\n".join(self.result)

    def cls(self):
        if platform.platform().startswith("Windows"):
            os.system("cls")
        else:
            print(chr(27)+"[2J"+chr(27)+"[1;1H", end="")  # ansi clear screen


if __name__ == "__main__":
    start = -2.0-1.0j
    width = 3.0
    duration = 30.0
    wallclock_start = time.time()
    frames = 0
    zoomer = MandelZoomer()
    zoomer.cls()
    print("This is a mandelbrot zoom animation running using Pyro, it will use all calculation server processes that are available.")
    while True:
        time_passed = time.time() - wallclock_start
        if time_passed >= duration:
            break
        actual_width = width * (1-time_passed/duration/1.1)
        actual_start = start + (0.06-0.002j)*time_passed
        frame = zoomer.screen(actual_start, actual_width)
        zoomer.cls()
        fps = frames/time_passed if time_passed > 0 else 0
        print("%.1f FPS time=%.2f width=%.2f" % (fps, time_passed, actual_width))
        print(frame)
        frames += 1
    print("Final FPS: %.2f" % fps)




import sys
import Pyro5.api
import Pyro5.errors

test = Pyro5.api.Proxy("PYRONAME:example.exceptions")

print(test.div(2.0, 9.0))
try:
    print(2 // 0)
except ZeroDivisionError as x:
    print("DIVIDE BY ZERO: ", x)
try:
    print(test.div(2, 0))
except ZeroDivisionError as x:
    print("DIVIDE BY ZERO: ", x)
try:
    result = test.error()
    print("%r, %s" % (result, result))
except ValueError as x:
    print("VALUERROR: ", x)
try:
    result = test.error2()
    print("%r, %s" % (result, result))
except ValueError as x:
    print("VALUERROR: ", x)
try:
    result = test.othererr()
    print("%r, %s" % (result, result))
except Exception as x:
    print("ANOTHER ERROR: ", x)
try:
    result = test.onewayerr()
    print("oneway call simply succeeded")
except Exception as x:
    print("SHOULD NOT HAPPEN: exception from oneway call", x)
try:
    result = test.unserializable()
    print("%r, %s" % (result, result))
except Exception as x:
    print("UNSERIALIZABLE ERROR: ", x)

print("\n*** invoking server method that crashes, catching traceback ***")
try:
    print(test.complexerror())
except Exception as x:
    print("CAUGHT ERROR  >>> ", x)
    print("Printing Pyro traceback >>>>>>")
    print("".join(Pyro5.errors.get_pyro_traceback()))
    print("<<<<<<< end of Pyro traceback")

print("\n*** installing pyro's excepthook")
sys.excepthook = Pyro5.errors.excepthook
print("*** invoking server method that crashes, not catching anything ***")
print(test.complexerror())  # due to the excepthook, the exception will show the pyro error



import Pyro5.api
import Pyro5.core
import Pyro5.client
import Pyro5.server
import Pyro5.nameserver
import Pyro5.callcontext
import Pyro5.serializers
from Pyro5.serializers import SerializerBase


def test_api():
    assert hasattr(Pyro5.api, "__version__")
    assert Pyro5.api.config.SERIALIZER == "serpent"
    assert Pyro5.api.URI is Pyro5.core.URI
    assert Pyro5.api.Proxy is Pyro5.client.Proxy
    assert Pyro5.api.Daemon is Pyro5.server.Daemon
    assert Pyro5.api.start_ns is Pyro5.nameserver.start_ns
    assert Pyro5.api.current_context is Pyro5.callcontext.current_context
    assert Pyro5.api.register_dict_to_class == SerializerBase.register_dict_to_class
    assert Pyro5.api.register_class_to_dict == SerializerBase.register_class_to_dict
    assert Pyro5.api.unregister_dict_to_class == SerializerBase.unregister_dict_to_class
    assert Pyro5.api.unregister_class_to_dict == SerializerBase.unregister_class_to_dict



from Pyro5.api import expose, behavior, serve


@expose
@behavior(instance_mode="single")
class Warehouse(object):
    def __init__(self):
        self.contents = ["chair", "bike", "flashlight", "laptop", "couch"]

    def list_contents(self):
        return self.contents

    def take(self, name, item):
        self.contents.remove(item)
        print("{0} took the {1}.".format(name, item))

    def store(self, name, item):
        self.contents.append(item)
        print("{0} stored the {1}.".format(name, item))


serve(
    {
        Warehouse: "example.warehouse"
    },
    use_ns=True)



"""
Client related classes (Proxy, mostly)

Pyro - Python Remote Objects.  Copyright by Irmen de Jong (irmen@razorvine.net).
"""

import sys
import time
import logging
import serpent
import contextlib
from . import config, core, serializers, protocol, errors, socketutil
from .callcontext import current_context
try:
    from greenlet import getcurrent as get_ident
except ImportError:
    from threading import get_ident


log = logging.getLogger("Pyro5.client")

__all__ = ["Proxy", "BatchProxy", "SerializedBlob"]


class Proxy(object):
    """
    Pyro proxy for a remote object. Intercepts method calls and dispatches them to the remote object.

    .. automethod:: _pyroBind
    .. automethod:: _pyroRelease
    .. automethod:: _pyroReconnect
    .. automethod:: _pyroValidateHandshake
    .. autoattribute:: _pyroTimeout
    .. attribute:: _pyroMaxRetries

        Number of retries to perform on communication calls by this proxy, allows you to override the default setting.

    .. attribute:: _pyroSerializer

        Name of the serializer to use by this proxy, allows you to override the default setting.

    .. attribute:: _pyroHandshake

        The data object that should be sent in the initial connection handshake message. Can be any serializable object.

    .. attribute:: _pyroLocalSocket

        The socket that is used locally to connect to the remote daemon.
        The format depends on the address family used for the connection, but usually
        for IPV4 connections it is the familiar (hostname, port) tuple.
        Consult the Python documentation on `socket families <https://docs.python.org/3/library/socket.html#socket-families>`_
        for more details
    """
    __pyroAttributes = frozenset(
        ["__getnewargs__", "__getnewargs_ex__", "__getinitargs__", "_pyroConnection", "_pyroUri",
         "_pyroOneway", "_pyroMethods", "_pyroAttrs", "_pyroTimeout", "_pyroSeq", "_pyroLocalSocket",
         "_pyroRawWireResponse", "_pyroHandshake", "_pyroMaxRetries", "_pyroSerializer",
         "_Proxy__pyroTimeout", "_Proxy__pyroOwnerThread"])

    def __init__(self, uri, connected_socket=None):
        if connected_socket:
            uri = core.URI("PYRO:" + uri + "@<<connected-socket>>:0")
        if isinstance(uri, str):
            uri = core.URI(uri)
        elif not isinstance(uri, core.URI):
            raise TypeError("expected Pyro URI")
        self._pyroUri = uri
        self._pyroConnection = None
        self._pyroSerializer = None  # can be set to the name of a serializer to override the global one per-proxy
        self._pyroMethods = set()  # all methods of the remote object, gotten from meta-data
        self._pyroAttrs = set()  # attributes of the remote object, gotten from meta-data
        self._pyroOneway = set()  # oneway-methods of the remote object, gotten from meta-data
        self._pyroSeq = 0  # message sequence number
        self._pyroRawWireResponse = False  # internal switch to enable wire level responses
        self._pyroHandshake = "hello"  # the data object that should be sent in the initial connection handshake message
        self._pyroMaxRetries = config.MAX_RETRIES
        self.__pyroTimeout = config.COMMTIMEOUT
        self.__pyroOwnerThread = get_ident()     # the thread that owns this proxy
        if config.SERIALIZER not in serializers.serializers:
            raise ValueError("unknown serializer configured")
        # note: we're not clearing the client annotations dict here.
        #       that is because otherwise it will be wiped if a new proxy is needed to connect PYRONAME uris.
        #       clearing the response annotations is okay.
        current_context.response_annotations = {}
        if connected_socket:
            self.__pyroCreateConnection(False, connected_socket)

    def __del__(self):
        if hasattr(self, "_pyroConnection"):
            try:
                self._pyroRelease()
            except Exception:
                pass

    def __getattr__(self, name):
        if name in Proxy.__pyroAttributes:
            # allows it to be safely pickled
            raise AttributeError(name)
        # get metadata if it's not there yet
        if not self._pyroMethods and not self._pyroAttrs:
            self._pyroGetMetadata()
        if name in self._pyroAttrs:
            return self._pyroInvoke("__getattr__", (name,), None)
        if name not in self._pyroMethods:
            # client side check if the requested attr actually exists
            raise AttributeError("remote object '%s' has no exposed attribute or method '%s'" % (self._pyroUri, name))
        return _RemoteMethod(self._pyroInvoke, name, self._pyroMaxRetries)

    def __setattr__(self, name, value):
        if name in Proxy.__pyroAttributes:
            return super(Proxy, self).__setattr__(name, value)  # one of the special pyro attributes
        # get metadata if it's not there yet
        if not self._pyroMethods and not self._pyroAttrs:
            self._pyroGetMetadata()
        if name in self._pyroAttrs:
            return self._pyroInvoke("__setattr__", (name, value), None)  # remote attribute
        # client side validation if the requested attr actually exists
        raise AttributeError("remote object '%s' has no exposed attribute '%s'" % (self._pyroUri, name))

    def __repr__(self):
        if self._pyroConnection:
            connected = "connected " + self._pyroConnection.family()
        else:
            connected = "not connected"
        return "<%s.%s at 0x%x; %s; for %s; owner %s>" % (self.__class__.__module__, self.__class__.__name__,
                                                          id(self), connected, self._pyroUri, self.__pyroOwnerThread)

    def __getstate__(self):
        # make sure a tuple of just primitive types are used to allow for proper serialization
        return str(self._pyroUri), tuple(self._pyroOneway), tuple(self._pyroMethods), \
               tuple(self._pyroAttrs), self._pyroHandshake, self._pyroSerializer

    def __setstate__(self, state):
        self._pyroUri = core.URI(state[0])
        self._pyroOneway = set(state[1])
        self._pyroMethods = set(state[2])
        self._pyroAttrs = set(state[3])
        self._pyroHandshake = state[4]
        self._pyroSerializer = state[5]
        self.__pyroTimeout = config.COMMTIMEOUT
        self._pyroMaxRetries = config.MAX_RETRIES
        self._pyroConnection = None
        self._pyroLocalSocket = None
        self._pyroSeq = 0
        self._pyroRawWireResponse = False
        self.__pyroOwnerThread = get_ident()

    def __copy__(self):
        p = object.__new__(type(self))
        p.__setstate__(self.__getstate__())
        p._pyroTimeout = self._pyroTimeout
        p._pyroRawWireResponse = self._pyroRawWireResponse
        p._pyroMaxRetries = self._pyroMaxRetries
        return p

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self._pyroRelease()

    def __eq__(self, other):
        if other is self:
            return True
        return isinstance(other, Proxy) and other._pyroUri == self._pyroUri

    def __ne__(self, other):
        if other and isinstance(other, Proxy):
            return other._pyroUri != self._pyroUri
        return True

    def __hash__(self):
        return hash(self._pyroUri)

    def __dir__(self):
        result = dir(self.__class__) + list(self.__dict__.keys())
        return sorted(set(result) | self._pyroMethods | self._pyroAttrs)

    # When special methods are invoked via special syntax (e.g. obj[index] calls
    # obj.__getitem__(index)), the special methods are not looked up via __getattr__
    # for efficiency reasons; instead, their presence is checked directly.
    # Thus we need to define them here to force (remote) lookup through __getitem__.
    def __bool__(self): return True
    def __len__(self): return self.__getattr__('__len__')()
    def __getitem__(self, index): return self.__getattr__('__getitem__')(index)
    def __setitem__(self, index, val): return self.__getattr__('__setitem__')(index, val)
    def __delitem__(self, index): return self.__getattr__('__delitem__')(index)

    def __iter__(self):
        try:
            # use remote iterator if it exists
            yield from self.__getattr__('__iter__')()
        except AttributeError:
            # fallback to indexed based iteration
            try:
                yield from (self[index] for index in range(sys.maxsize))
            except (StopIteration, IndexError):
                return

    def _pyroRelease(self):
        """release the connection to the pyro daemon"""
        self.__check_owner()
        if self._pyroConnection is not None:
            self._pyroConnection.close()
            self._pyroConnection = None
            self._pyroLocalSocket = None

    def _pyroBind(self):
        """
        Bind this proxy to the exact object from the uri. That means that the proxy's uri
        will be updated with a direct PYRO uri, if it isn't one yet.
        If the proxy is already bound, it will not bind again.
        """
        return self.__pyroCreateConnection(True)

    def __pyroGetTimeout(self):
        return self.__pyroTimeout

    def __pyroSetTimeout(self, timeout):
        self.__pyroTimeout = timeout
        if self._pyroConnection is not None:
            self._pyroConnection.timeout = timeout

    _pyroTimeout = property(__pyroGetTimeout, __pyroSetTimeout, doc="""
        The timeout in seconds for calls on this proxy. Defaults to ``None``.
        If the timeout expires before the remote method call returns,
        Pyro will raise a :exc:`Pyro5.errors.TimeoutError`""")

    def _pyroInvoke(self, methodname, vargs, kwargs, flags=0, objectId=None):
        """perform the remote method call communication"""
        self.__check_owner()
        current_context.response_annotations = {}
        if self._pyroConnection is None:
            self.__pyroCreateConnection()
        serializer = serializers.serializers[self._pyroSerializer or config.SERIALIZER]
        objectId = objectId or self._pyroConnection.objectId
        annotations = current_context.annotations
        if vargs and isinstance(vargs[0], SerializedBlob):
            # special serialization of a 'blob' that stays serialized
            data, flags = self.__serializeBlobArgs(vargs, kwargs, annotations, flags, objectId, methodname, serializer)
        else:
            # normal serialization of the remote call
            data = serializer.dumpsCall(objectId, methodname, vargs, kwargs)
        if methodname in self._pyroOneway:
            flags |= protocol.FLAGS_ONEWAY
        self._pyroSeq = (self._pyroSeq + 1) & 0xffff
        msg = protocol.SendingMessage(protocol.MSG_INVOKE, flags, self._pyroSeq, serializer.serializer_id, data, annotations=annotations)
        if config.LOGWIRE:
            protocol.log_wiredata(log, "proxy wiredata sending", msg)
        try:
            self._pyroConnection.send(msg.data)
            del msg  # invite GC to collect the object, don't wait for out-of-scope
            if flags & protocol.FLAGS_ONEWAY:
                return None  # oneway call, no response data
            else:
                msg = protocol.recv_stub(self._pyroConnection, [protocol.MSG_RESULT])
                if config.LOGWIRE:
                    protocol.log_wiredata(log, "proxy wiredata received", msg)
                self.__pyroCheckSequence(msg.seq)
                if msg.serializer_id != serializer.serializer_id:
                    error = "invalid serializer in response: %d" % msg.serializer_id
                    log.error(error)
                    raise errors.SerializeError(error)
                if msg.annotations:
                    current_context.response_annotations = msg.annotations
                if self._pyroRawWireResponse:
                    return msg
                data = serializer.loads(msg.data)
                if msg.flags & protocol.FLAGS_ITEMSTREAMRESULT:
                    streamId = bytes(msg.annotations.get("STRM", b"")).decode()
                    if not streamId:
                        raise errors.ProtocolError("result of call is an iterator, but the server is not configured to allow streaming")
                    return _StreamResultIterator(streamId, self)
                if msg.flags & protocol.FLAGS_EXCEPTION:
                    raise data  # if you see this in your traceback, you should probably inspect the remote traceback as well
                else:
                    return data
        except (errors.CommunicationError, KeyboardInterrupt):
            # Communication error during read. To avoid corrupt transfers, we close the connection.
            # Otherwise we might receive the previous reply as a result of a new method call!
            # Special case for keyboardinterrupt: people pressing ^C to abort the client
            # may be catching the keyboardinterrupt in their code. We should probably be on the
            # safe side and release the proxy connection in this case too, because they might
            # be reusing the proxy object after catching the exception...
            self._pyroRelease()
            raise

    def __pyroCheckSequence(self, seq):
        if seq != self._pyroSeq:
            err = "invoke: reply sequence out of sync, got %d expected %d" % (seq, self._pyroSeq)
            log.error(err)
            raise errors.ProtocolError(err)

    def __pyroCreateConnection(self, replaceUri=False, connected_socket=None):
        """
        Connects this proxy to the remote Pyro daemon. Does connection handshake.
        Returns true if a new connection was made, false if an existing one was already present.
        """
        def connect_and_handshake(conn):
            try:
                if self._pyroConnection is not None:
                    return False  # already connected
                if config.SSL:
                    sslContext = socketutil.get_ssl_context(clientcert=config.SSL_CLIENTCERT,
                                                            clientkey=config.SSL_CLIENTKEY,
                                                            keypassword=config.SSL_CLIENTKEYPASSWD,
                                                            cacerts=config.SSL_CACERTS)
                else:
                    sslContext = None
                sock = socketutil.create_socket(connect=connect_location,
                                                reuseaddr=config.SOCK_REUSE,
                                                timeout=self.__pyroTimeout,
                                                nodelay=config.SOCK_NODELAY,
                                                sslContext=sslContext)
                conn = socketutil.SocketConnection(sock, uri.object)
                # Do handshake.
                serializer = serializers.serializers[self._pyroSerializer or config.SERIALIZER]
                data = {"handshake": self._pyroHandshake, "object": uri.object}
                data = serializer.dumps(data)
                msg = protocol.SendingMessage(protocol.MSG_CONNECT, 0, self._pyroSeq, serializer.serializer_id,
                                              data, annotations=current_context.annotations)
                if config.LOGWIRE:
                    protocol.log_wiredata(log, "proxy connect sending", msg)
                conn.send(msg.data)
                msg = protocol.recv_stub(conn, [protocol.MSG_CONNECTOK, protocol.MSG_CONNECTFAIL])
                if config.LOGWIRE:
                    protocol.log_wiredata(log, "proxy connect response received", msg)
            except Exception as x:
                if conn:
                    conn.close()
                err = "cannot connect to %s: %s" % (connect_location, x)
                log.error(err)
                if isinstance(x, errors.CommunicationError):
                    raise
                else:
                    raise errors.CommunicationError(err) from x
            else:
                handshake_response = "?"
                if msg.data:
                    serializer = serializers.serializers_by_id[msg.serializer_id]
                    handshake_response = serializer.loads(msg.data)
                if msg.type == protocol.MSG_CONNECTFAIL:
                    error = "connection to %s rejected: %s" % (connect_location, handshake_response)
                    conn.close()
                    log.error(error)
                    raise errors.CommunicationError(error)
                elif msg.type == protocol.MSG_CONNECTOK:
                    self.__processMetadata(handshake_response["meta"])
                    handshake_response = handshake_response["handshake"]
                    self._pyroConnection = conn
                    self._pyroLocalSocket = conn.sock.getsockname()
                    if replaceUri:
                        self._pyroUri = uri
                    self._pyroValidateHandshake(handshake_response)
                    log.debug("connected to %s - %s - %s", self._pyroUri, conn.family(), "SSL" if sslContext else "unencrypted")
                    if msg.annotations:
                        current_context.response_annotations = msg.annotations
                else:
                    conn.close()
                    err = "cannot connect to %s: invalid msg type %d received" % (connect_location, msg.type)
                    log.error(err)
                    raise errors.ProtocolError(err)

        self.__check_owner()
        if self._pyroConnection is not None:
            return False  # already connected
        uri = core.resolve(self._pyroUri)
        # socket connection (normal or Unix domain socket)
        conn = None
        log.debug("connecting to %s", uri)
        connect_location = uri.sockname or (uri.host, uri.port)
        if connected_socket:
            self._pyroConnection = socketutil.SocketConnection(connected_socket, uri.object, True)
            self._pyroLocalSocket = connected_socket.getsockname()
        else:
            connect_and_handshake(conn)
        # obtain metadata if this feature is enabled, and the metadata is not known yet
        if not self._pyroMethods and not self._pyroAttrs:
            self._pyroGetMetadata(uri.object)
        return True

    def _pyroGetMetadata(self, objectId=None, known_metadata=None):
        """
        Get metadata from server (methods, attrs, oneway, ...) and remember them in some attributes of the proxy.
        Usually this will already be known due to the default behavior of the connect handshake, where the
        connect response also includes the metadata.
        """
        objectId = objectId or self._pyroUri.object
        log.debug("getting metadata for object %s", objectId)
        if self._pyroConnection is None and not known_metadata:
            try:
                self.__pyroCreateConnection()
            except errors.PyroError:
                log.error("problem getting metadata: cannot connect")
                raise
            if self._pyroMethods or self._pyroAttrs:
                return  # metadata has already been retrieved as part of creating the connection
        try:
            # invoke the get_metadata method on the daemon
            result = known_metadata or self._pyroInvoke("get_metadata", [objectId], {}, objectId=core.DAEMON_NAME)
            self.__processMetadata(result)
        except errors.PyroError:
            log.exception("problem getting metadata")
            raise

    def __processMetadata(self, metadata):
        if not metadata:
            return
        self._pyroOneway = set(metadata["oneway"])
        self._pyroMethods = set(metadata["methods"])
        self._pyroAttrs = set(metadata["attrs"])
        if log.isEnabledFor(logging.DEBUG):
            log.debug("from meta: methods=%s, oneway methods=%s, attributes=%s",
                      sorted(self._pyroMethods), sorted(self._pyroOneway), sorted(self._pyroAttrs))
        if not self._pyroMethods and not self._pyroAttrs:
            raise errors.PyroError("remote object '%s' doesn't expose any methods or attributes. Did you forget setting @expose on them?" % self._pyroUri)

    def _pyroReconnect(self, tries=100000000):
        """
        (Re)connect the proxy to the daemon containing the pyro object which the proxy is for.
        In contrast to the _pyroBind method, this one first releases the connection (if the proxy is still connected)
        and retries making a new connection until it succeeds or the given amount of tries ran out.
        """
        self._pyroRelease()
        while tries:
            try:
                self.__pyroCreateConnection()
                return
            except errors.CommunicationError:
                tries -= 1
                if tries:
                    time.sleep(2)
        msg = "failed to reconnect"
        log.error(msg)
        raise errors.ConnectionClosedError(msg)

    def _pyroInvokeBatch(self, calls, oneway=False):
        flags = protocol.FLAGS_BATCH
        if oneway:
            flags |= protocol.FLAGS_ONEWAY
        return self._pyroInvoke("<batch>", calls, None, flags)

    def _pyroValidateHandshake(self, response):
        """
        Process and validate the initial connection handshake response data received from the daemon.
        Simply return without error if everything is ok.
        Raise an exception if something is wrong and the connection should not be made.
        """
        return

    def _pyroClaimOwnership(self):
        """
        The current thread claims the ownership of this proxy from another thread.
        Any existing connection will remain active!
        """
        if get_ident() != self.__pyroOwnerThread:
            # if self._pyroConnection is not None:
            #     self._pyroConnection.close()
            #     self._pyroConnection = None
            self.__pyroOwnerThread = get_ident()

    def __serializeBlobArgs(self, vargs, kwargs, annotations, flags, objectId, methodname, serializer):
        """
        Special handling of a "blob" argument that has to stay serialized until explicitly deserialized in client code.
        This makes efficient, transparent gateways or dispatchers and such possible:
        they don't have to de/reserialize the message and are independent from the serialized class definitions.
        Annotations are passed in because some blob metadata is added. They're not part of the blob itself.
        """
        if len(vargs) > 1 or kwargs:
            raise errors.SerializeError("if SerializedBlob is used, it must be the only argument")
        blob = vargs[0]
        flags |= protocol.FLAGS_KEEPSERIALIZED
        # Pass the objectId and methodname separately in an annotation because currently,
        # they are embedded inside the serialized message data. And we're not deserializing that,
        # so we have to have another means of knowing the object and method it is meant for...
        # A better solution is perhaps to split the actual remote method arguments from the
        # control data (object + methodname) but that requires a major protocol change.
        # The code below is not as nice but it works without any protocol change and doesn't
        # require a hack either - so it's actually not bad like this.
        import marshal
        annotations["BLBI"] = marshal.dumps((blob.info, objectId, methodname))
        if blob._contains_blob:
            # directly pass through the already serialized msg data from within the blob
            protocol_msg = blob._data
            return protocol_msg.data, flags
        else:
            # replaces SerializedBlob argument with the data to be serialized
            return serializer.dumpsCall(objectId, methodname, blob._data, kwargs), flags

    def __check_owner(self):
        if get_ident() != self.__pyroOwnerThread:
            raise errors.PyroError("the calling thread is not the owner of this proxy, "
                                   "create a new proxy in this thread or transfer ownership.")


class _RemoteMethod(object):
    """method call abstraction"""

    def __init__(self, send, name, max_retries):
        self.__send = send
        self.__name = name
        self.__max_retries = max_retries

    def __getattr__(self, name):
        return _RemoteMethod(self.__send, "%s.%s" % (self.__name, name), self.__max_retries)

    def __call__(self, *args, **kwargs):
        for attempt in range(self.__max_retries + 1):
            try:
                return self.__send(self.__name, args, kwargs)
            except (errors.ConnectionClosedError, errors.TimeoutError):
                # only retry for recoverable network errors
                if attempt >= self.__max_retries:
                    # last attempt, raise the exception
                    raise


class _StreamResultIterator(object):
    """
    Pyro returns this as a result of a remote call which returns an iterator or generator.
    It is a normal iterable and produces elements on demand from the remote iterator.
    You can simply use it in for loops, list comprehensions etc.
    """
    def __init__(self, streamId, proxy):
        self.streamId = streamId
        self.proxy = proxy
        self.pyroseq = proxy._pyroSeq

    def __iter__(self):
        return self

    def __next__(self):
        if self.proxy is None:
            raise StopIteration
        if self.proxy._pyroConnection is None:
            raise errors.ConnectionClosedError("the proxy for this stream result has been closed")
        self.pyroseq += 1
        try:
            return self.proxy._pyroInvoke("get_next_stream_item", [self.streamId], {}, objectId=core.DAEMON_NAME)
        except (StopIteration, GeneratorExit):
            # when the iterator is exhausted, the proxy is removed to avoid unneeded close_stream calls later
            # (the server has closed its part of the stream by itself already)
            self.proxy = None
            raise

    def __del__(self):
        try:
            self.close()
        except Exception:
            pass

    def close(self):
        if self.proxy and self.proxy._pyroConnection is not None:
            if self.pyroseq == self.proxy._pyroSeq:
                # we're still in sync, it's okay to use the same proxy to close this stream
                self.proxy._pyroInvoke("close_stream", [self.streamId], {},
                                       flags=protocol.FLAGS_ONEWAY, objectId=core.DAEMON_NAME)
            else:
                # The proxy's sequence number has diverged.
                # One of the reasons this can happen is because this call is being done from python's GC where
                # it decides to gc old iterator objects *during a new call on the proxy*.
                # If we use the same proxy and do a call in between, the other call on the proxy will get an out of sync seq and crash!
                # We create a temporary second proxy to call close_stream on. This is inefficient, but avoids the problem.
                with contextlib.suppress(errors.CommunicationError):
                    with self.proxy.__copy__() as closingProxy:
                        closingProxy._pyroInvoke("close_stream", [self.streamId], {},
                                                 flags=protocol.FLAGS_ONEWAY, objectId=core.DAEMON_NAME)
        self.proxy = None


class _BatchedRemoteMethod(object):
    """method call abstraction that is used with batched calls"""

    def __init__(self, calls, name):
        self.__calls = calls
        self.__name = name

    def __getattr__(self, name):
        return _BatchedRemoteMethod(self.__calls, "%s.%s" % (self.__name, name))

    def __call__(self, *args, **kwargs):
        self.__calls.append((self.__name, args, kwargs))


class BatchProxy(object):
    """Proxy that lets you batch multiple method calls into one.
    It is constructed with a reference to the normal proxy that will
    carry out the batched calls. Call methods on this object that you want to batch,
    and finally call the batch proxy itself. That call will return a generator
    for the results of every method call in the batch (in sequence)."""

    def __init__(self, proxy):
        self.__proxy = proxy
        self.__calls = []

    def __getattr__(self, name):
        return _BatchedRemoteMethod(self.__calls, name)

    def __enter__(self):
        return self

    def __exit__(self, *args):
        pass

    def __copy__(self):
        copy = type(self)(self.__proxy)
        copy.__calls = list(self.__calls)
        return copy

    def __resultsgenerator(self, results):
        for result in results:
            if isinstance(result, core._ExceptionWrapper):
                result.raiseIt()  # re-raise the remote exception locally.
            else:
                yield result  # it is a regular result object, yield that and continue.

    def __call__(self, oneway=False):
        self.__proxy._pyroClaimOwnership()
        results = self.__proxy._pyroInvokeBatch(self.__calls, oneway)
        self.__calls = []  # clear for re-use
        if not oneway:
            return self.__resultsgenerator(results)

    def _pyroInvoke(self, name, args, kwargs):
        # ignore all parameters, we just need to execute the batch
        results = self.__proxy._pyroInvokeBatch(self.__calls)
        self.__calls = []  # clear for re-use
        return self.__resultsgenerator(results)


class SerializedBlob(object):
    """
    Used to wrap some data to make Pyro pass this object transparently (it keeps the serialized payload as-is)
    Only when you need to access the actual client data you can deserialize on demand.
    This makes efficient, transparent gateways or dispatchers and such possible:
    they don't have to de/reserialize the message and are independent from the serialized class definitions.
    You have to pass this as the only parameter to a remote method call for Pyro to understand it.
    Init arguments:
    ``info`` = some (small) descriptive data about the blob. Can be a simple id or name or guid. Must be marshallable.
    ``data`` = the actual client data payload that you want to transfer in the blob. Can be anything that you would
    otherwise have used as regular remote call arguments.
    """
    def __init__(self, info, data, is_blob=False):
        self.info = info
        self._data = data
        self._contains_blob = is_blob
        if is_blob and not isinstance(data, (protocol.SendingMessage, protocol.ReceivingMessage)):
            raise TypeError("data should be a protocol message object if is_blob is true")

    def deserialized(self):
        """Retrieves the client data stored in this blob. Deserializes the data automatically if required."""
        if self._contains_blob:
            protocol_msg = self._data
            serializer = serializers.serializers_by_id[protocol_msg.serializer_id]
            if isinstance(protocol_msg, protocol.ReceivingMessage):
                _, _, data, _ = serializer.loads(protocol_msg.data)
            else:
                # strip off header bytes from SendingMessage
                payload_data = memoryview(protocol_msg.data)[protocol._header_size:]
                _, _, data, _ = serializer.loads(payload_data)
            return data
        else:
            return self._data


# register the special serializers for the pyro objects
serpent.register_class(Proxy, serializers.pyro_class_serpent_serializer)
serializers.SerializerBase.register_class_to_dict(Proxy, serializers.serialize_pyro_object_to_dict, serpent_too=False)



import Pyro5.core
import Pyro5.api

secret_code = "pancakes"


class CustomDaemon(Pyro5.api.Daemon):
    def validateHandshake(self, conn, data):
        print("Daemon received handshake request from:", conn.sock.getpeername())
        print("Handshake data:", data)

        # if needed, you can inspect Pyro5.callcontext.current_context:
        ctx = Pyro5.api.current_context
        print("  context.client: ", ctx.client)
        print("  context.client_sock_addr: ", ctx.client_sock_addr)
        print("  context.seq: ", ctx.seq)
        print("  context.msg_flags: ", ctx.msg_flags)
        print("  context.serializer_id: ", ctx.serializer_id)
        print("  context.correlation_id:", ctx.correlation_id)

        if data == secret_code:
            print("Secret code okay! Connection accepted.")
            # return some custom handshake data:
            return ["how", "are", "you", "doing"]
        else:
            print("Secret code wrong! Connection refused.")
            raise ValueError("wrong secret code, connection refused")

    def clientDisconnect(self, conn):
        print("Daemon client disconnects:", conn.sock.getpeername())


with CustomDaemon() as daemon:
    print("Server is ready. You can use the following URI to connect:")
    print(daemon.uriFor(Pyro5.core.DAEMON_NAME))
    print("When asked, enter the following secret code: ", secret_code)
    daemon.requestLoop()



"""
Tests for the name server.

Pyro - Python Remote Objects.  Copyright by Irmen de Jong (irmen@razorvine.net).
"""

import time
import pytest
import threading
import os
import sys
from io import StringIO
import Pyro5.core
import Pyro5.client
import Pyro5.nsc
import Pyro5.nameserver
import Pyro5.socketutil
from Pyro5.errors import CommunicationError, NamingError, PyroError
from Pyro5 import config


class NSLoopThread(threading.Thread):
    def __init__(self, nameserver):
        super(NSLoopThread, self).__init__()
        self.daemon = True
        self.nameserver = nameserver
        self.running = threading.Event()
        self.running.clear()

    def run(self):
        self.running.set()
        try:
            self.nameserver.requestLoop()
        except CommunicationError:
            pass  # ignore pyro communication errors


class TestBCSetup:
    @pytest.mark.network
    def testBCstart(self):
        myIpAddress = Pyro5.socketutil.get_ip_address("", workaround127=True)
        nsUri, nameserver, bcserver = Pyro5.nameserver.start_ns(host=myIpAddress, port=0, bcport=0, enableBroadcast=False)
        assert bcserver is None
        nameserver.close()
        nsUri, nameserver, bcserver = Pyro5.nameserver.start_ns(host=myIpAddress, port=0, bcport=0, enableBroadcast=True)
        assert bcserver is not None
        assert bcserver.fileno() > 1
        assert bcserver.sock is not None
        nameserver.close()
        bcserver.close()


class TestNameServer:
    def setup_method(self):
        config.POLLTIMEOUT = 0.1
        myIpAddress = Pyro5.socketutil.get_ip_address("", workaround127=True)
        self.nsUri, self.nameserver, self.bcserver = Pyro5.nameserver.start_ns(host=myIpAddress, port=0, bcport=0)
        assert self.bcserver is not None
        self.bcserver.runInThread()
        self.daemonthread = NSLoopThread(self.nameserver)
        self.daemonthread.start()
        self.daemonthread.running.wait()
        time.sleep(0.05)
        self.old_bcPort = config.NS_BCPORT
        self.old_nsPort = config.NS_PORT
        self.old_nsHost = config.NS_HOST
        config.NS_PORT = self.nsUri.port
        config.NS_HOST = str(myIpAddress)
        config.NS_BCPORT = self.bcserver.getPort()

    def teardown_method(self):
        time.sleep(0.01)
        self.nameserver.shutdown()
        self.bcserver.close()
        self.daemonthread.join()
        config.NS_HOST = self.old_nsHost
        config.NS_PORT = self.old_nsPort
        config.NS_BCPORT = self.old_bcPort

    @pytest.mark.network
    def testLookupUnixsockParsing(self):
        # this must not raise AttributeError, it did before because of a parse bug
        with pytest.raises(NamingError):
            Pyro5.core.locate_ns("./u:/tmp/Pyro5-naming.usock")

    @pytest.mark.network
    def testLookupAndRegister(self):
        ns = Pyro5.core.locate_ns()  # broadcast lookup
        assert isinstance(ns, Pyro5.client.Proxy)
        ns._pyroRelease()
        ns = Pyro5.core.locate_ns(self.nsUri.host)  # normal lookup
        assert isinstance(ns, Pyro5.client.Proxy)
        uri = ns._pyroUri
        assert uri.protocol == "PYRO"
        assert uri.host == self.nsUri.host
        assert uri.port == config.NS_PORT
        ns._pyroRelease()
        ns = Pyro5.core.locate_ns(self.nsUri.host, config.NS_PORT)
        uri = ns._pyroUri
        assert uri.protocol == "PYRO"
        assert uri.host == self.nsUri.host
        assert uri.port == config.NS_PORT
        # check that we cannot register a stupid type
        with pytest.raises(TypeError):
            ns.register("unittest.object1", 5555)
        # we can register str or URI, lookup always returns URI
        ns.register("unittest.object2", "PYRO:55555@host.com:4444")
        assert ns.lookup("unittest.object2") == Pyro5.core.URI("PYRO:55555@host.com:4444")
        ns.register("unittest.object3", Pyro5.core.URI("PYRO:66666@host.com:4444"))
        assert ns.lookup("unittest.object3") == Pyro5.core.URI("PYRO:66666@host.com:4444")
        ns._pyroRelease()

    @pytest.mark.network
    def testDaemonPyroObj(self):
        uri = self.nsUri
        uri.object = Pyro5.core.DAEMON_NAME
        with Pyro5.client.Proxy(uri) as daemonobj:
            daemonobj.ping()
            daemonobj.registered()
            with pytest.raises(AttributeError):
                daemonobj.shutdown()

    @pytest.mark.network
    def testMulti(self):
        uristr = str(self.nsUri)
        p = Pyro5.client.Proxy(uristr)
        p._pyroBind()
        p._pyroRelease()
        uri = Pyro5.core.resolve(uristr)
        p = Pyro5.client.Proxy(uri)
        p._pyroBind()
        p._pyroRelease()
        uri = Pyro5.core.resolve(uristr)
        p = Pyro5.client.Proxy(uri)
        p._pyroBind()
        p._pyroRelease()
        uri = Pyro5.core.resolve(uristr)
        p = Pyro5.client.Proxy(uri)
        p._pyroBind()
        p._pyroRelease()
        uri = Pyro5.core.resolve(uristr)
        p = Pyro5.client.Proxy(uri)
        p._pyroBind()
        p._pyroRelease()
        uri = Pyro5.core.resolve(uristr)
        p = Pyro5.client.Proxy(uri)
        p._pyroBind()
        p._pyroRelease()
        daemonUri = "PYRO:" + Pyro5.core.DAEMON_NAME + "@" + uri.location
        _ = Pyro5.core.resolve(daemonUri)
        _ = Pyro5.core.resolve(daemonUri)
        _ = Pyro5.core.resolve(daemonUri)
        _ = Pyro5.core.resolve(daemonUri)
        _ = Pyro5.core.resolve(daemonUri)
        _ = Pyro5.core.resolve(daemonUri)
        uri = Pyro5.core.resolve(daemonUri)
        pyronameUri = "PYRONAME:" + Pyro5.core.NAMESERVER_NAME + "@" + uri.location
        _ = Pyro5.core.resolve(pyronameUri)
        _ = Pyro5.core.resolve(pyronameUri)
        _ = Pyro5.core.resolve(pyronameUri)
        _ = Pyro5.core.resolve(pyronameUri)
        _ = Pyro5.core.resolve(pyronameUri)
        _ = Pyro5.core.resolve(pyronameUri)

    @pytest.mark.network
    def testResolve(self):
        resolved1 = Pyro5.core.resolve(Pyro5.core.URI("PYRO:12345@host.com:4444"))
        resolved2 = Pyro5.core.resolve("PYRO:12345@host.com:4444")
        assert type(resolved1) is Pyro5.core.URI
        assert resolved1 == resolved2
        assert str(resolved1) == "PYRO:12345@host.com:4444"

        ns = Pyro5.core.locate_ns(self.nsUri.host, self.nsUri.port)
        host = "[" + self.nsUri.host + "]" if ":" in self.nsUri.host else self.nsUri.host
        uri = Pyro5.core.resolve("PYRONAME:" + Pyro5.core.NAMESERVER_NAME + "@" + host + ":" + str(self.nsUri.port))
        assert uri.protocol == "PYRO"
        assert uri.host == self.nsUri.host
        assert uri.object == Pyro5.core.NAMESERVER_NAME
        assert ns._pyroUri == uri
        ns._pyroRelease()

        # broadcast lookup
        with pytest.raises(NamingError):
            Pyro5.core.resolve("PYRONAME:unknown_object")
        uri = Pyro5.core.resolve("PYRONAME:" + Pyro5.core.NAMESERVER_NAME)
        assert isinstance(uri, Pyro5.core.URI)
        assert uri.protocol == "PYRO"

        # test some errors
        with pytest.raises(NamingError):
            Pyro5.core.resolve("PYRONAME:unknown_object@" + host)
        with pytest.raises(TypeError):
            Pyro5.core.resolve(999)

    @pytest.mark.network
    def testRefuseDottedNames(self):
        with Pyro5.core.locate_ns(self.nsUri.host, self.nsUri.port) as ns:
            # the name server should never have dotted names enabled
            with pytest.raises(AttributeError):
                _ = ns.namespace.keys
            assert ns._pyroConnection is not None
        assert ns._pyroConnection is None

    @pytest.mark.network
    def testAutoClean(self):
        try:
            config.NS_AUTOCLEAN = 0.0
            config.COMMTIMEOUT = 0.5
            Pyro5.nameserver.AutoCleaner.max_unreachable_time = 1
            Pyro5.nameserver.AutoCleaner.loop_delay = 0.5
            Pyro5.nameserver.AutoCleaner.override_autoclean_min = True
            with Pyro5.nameserver.NameServerDaemon(port=0) as ns:
                assert ns.cleaner_thread is None
            config.NS_AUTOCLEAN = 0.2
            with Pyro5.nameserver.NameServerDaemon(port=0) as ns:
                assert ns.cleaner_thread is not None
                ns.nameserver.register("test", "PYRO:test@localhost:59999")
                assert ns.nameserver.count() == 2
                time.sleep(4)
                assert ns.nameserver.count() == 1
            assert ns.cleaner_thread is None
        finally:
            Pyro5.nameserver.AutoCleaner.override_autoclean_min = False
            Pyro5.nameserver.AutoCleaner.max_unreachable_time = 20
            Pyro5.nameserver.AutoCleaner.loop_delay = 2
            config.NS_AUTOCLEAN = 0.0
            config.COMMTIMEOUT = 0.0


class TestNameServer0000:
    def setup_method(self):
        config.POLLTIMEOUT = 0.1
        self.nsUri, self.nameserver, self.bcserver = Pyro5.nameserver.start_ns(host="", port=0, bcport=0)
        host_check = self.nsUri.host
        assert host_check == "0.0.0.0"
        assert self.bcserver is not None
        self.bcthread = self.bcserver.runInThread()
        self.old_bcPort = config.NS_BCPORT
        self.old_nsPort = config.NS_PORT
        self.old_nsHost = config.NS_HOST
        config.NS_PORT = self.nsUri.port
        config.NS_HOST = self.nsUri.host
        config.NS_BCPORT = self.bcserver.getPort()

    def teardown_method(self):
        time.sleep(0.01)
        self.nameserver.shutdown()
        self.bcserver.close()
        self.bcthread.join()
        config.NS_HOST = self.old_nsHost
        config.NS_PORT = self.old_nsPort
        config.NS_BCPORT = self.old_bcPort

    @pytest.mark.network
    def testBCLookup0000(self):
        ns = Pyro5.core.locate_ns()  # broadcast lookup
        assert isinstance(ns, Pyro5.client.Proxy)
        assert ns._pyroUri.host != "0.0.0.0"
        ns._pyroRelease()


class TestOfflineNameServer:
    def setup_method(self):
        self.storageProvider = Pyro5.nameserver.MemoryStorage()

    def teardown_method(self):
        self.clearStorage()
        self.clearStorage()
        self.storageProvider.close()

    def clearStorage(self):
        try:
            self.storageProvider.clear()
        except AttributeError:
            pass

    def testRegister(self):
        ns = Pyro5.nameserver.NameServer(storageProvider=self.storageProvider)
        self.clearStorage()
        ns.ping()
        ns.register("test.object1", "PYRO:000000@host.com:4444")
        ns.register("test.object2", "PYRO:222222@host.com:4444")
        ns.register("test.object3", "PYRO:333333@host.com:4444")
        assert str(ns.lookup("test.object1")) == "PYRO:000000@host.com:4444"
        ns.register("test.object1", "PYRO:111111@host.com:4444")  # registering again should be ok by default
        assert str(ns.lookup("test.object1")), "should be new uri" == "PYRO:111111@host.com:4444"
        ns.register("test.sub.objectA", Pyro5.core.URI("PYRO:AAAAAA@host.com:4444"))
        ns.register("test.sub.objectB", Pyro5.core.URI("PYRO:BBBBBB@host.com:4444"))

        # if safe=True, a registration of an existing name should give a NamingError
        with pytest.raises(NamingError):
            ns.register("test.object1", "PYRO:X@Y:5555", safe=True)
        with pytest.raises(TypeError):
            ns.register(None, None)
        with pytest.raises(TypeError):
            ns.register(4444, 4444)
        with pytest.raises(TypeError):
            ns.register("test.wrongtype", 4444)
        with pytest.raises(TypeError):
            ns.register(4444, "PYRO:X@Y:5555")
        with pytest.raises(NamingError):
            ns.lookup("unknown_object")

        uri = ns.lookup("test.object3")
        assert uri == Pyro5.core.URI("PYRO:333333@host.com:4444")  # lookup always returns URI
        ns.remove("unknown_object")
        ns.remove("test.object1")
        ns.remove("test.object2")
        ns.remove("test.object3")
        all_objs = ns.list()
        assert len(all_objs) == 2  # 2 leftover objects
        with pytest.raises(PyroError):
            ns.register("test.nonurivalue", "THISVALUEISNOTANURI")
        ns.storage.close()

    def testRemove(self):
        ns = Pyro5.nameserver.NameServer(storageProvider=self.storageProvider)
        self.clearStorage()
        ns.register(Pyro5.core.NAMESERVER_NAME, "PYRO:nameserver@host:555")
        for i in range(20):
            ns.register("test.%d" % i, "PYRO:obj@host:555")
        assert len(ns.list()) == 21
        assert ns.remove("wrong") == 0
        assert ns.remove(prefix="wrong") == 0
        assert ns.remove(regex="wrong.*") == 0
        assert ns.remove("test.0") == 1
        assert len(ns.list()) == 20
        assert ns.remove(prefix="test.1") == 11  # 1, 10-19
        assert ns.remove(regex=r"test\..") == 8  # 2-9
        assert len(ns.list()) == 1
        ns.storage.close()

    def testRemoveProtected(self):
        ns = Pyro5.nameserver.NameServer(storageProvider=self.storageProvider)
        self.clearStorage()
        ns.register(Pyro5.core.NAMESERVER_NAME, "PYRO:nameserver@host:555")
        assert ns.remove(Pyro5.core.NAMESERVER_NAME) == 0
        assert ns.remove(prefix="Pyro") == 0
        assert ns.remove(regex="Pyro.*") == 0
        assert Pyro5.core.NAMESERVER_NAME in ns.list()
        ns.storage.close()

    def testUnicodeNames(self):
        ns = Pyro5.nameserver.NameServer(storageProvider=self.storageProvider)
        self.clearStorage()
        uri = Pyro5.core.URI("PYRO:unicode" + chr(0x20ac) + "@host:5555")
        ns.register("unicodename" + chr(0x20ac), uri)
        x = ns.lookup("unicodename" + chr(0x20ac))
        assert x == uri
        ns.storage.close()

    def testList(self):
        ns = Pyro5.nameserver.NameServer(storageProvider=self.storageProvider)
        self.clearStorage()
        ns.register("test.objects.1", "PYRONAME:something1")
        ns.register("test.objects.2", "PYRONAME:something2")
        ns.register("test.objects.3", "PYRONAME:something3")
        ns.register("test.other.a", "PYRONAME:somethingA")
        ns.register("test.other.b", "PYRONAME:somethingB")
        ns.register("test.other.c", "PYRONAME:somethingC")
        ns.register("entirely.else", "PYRONAME:meh")
        objects = ns.list()
        assert len(objects) == 7
        objects = ns.list(prefix="nothing")
        assert len(objects) == 0
        objects = ns.list(prefix="test.")
        assert len(objects) == 6
        objects = ns.list(regex=r".+other..")
        assert len(objects) == 3
        assert "test.other.a" in objects
        assert objects["test.other.a"] == "PYRONAME:somethingA"
        objects = ns.list(regex=r"\d\d\d\d\d\d\d\d\d\d")
        assert len(objects) == 0
        with pytest.raises(NamingError):
            ns.list(regex="((((((broken")
        ns.storage.close()

    def testNameserverWithStmt(self):
        ns = Pyro5.nameserver.NameServerDaemon(port=0)
        assert ns.nameserver is not None
        ns.close()
        assert ns.nameserver is None
        with Pyro5.nameserver.NameServerDaemon(port=0) as ns:
            assert ns.nameserver is not None
            pass
        assert ns.nameserver is None
        with pytest.raises(ZeroDivisionError):
            with Pyro5.nameserver.NameServerDaemon(port=0) as ns:
                assert ns.nameserver is not None
                print(1 // 0)  # cause an error
        assert ns.nameserver is None
        ns = Pyro5.nameserver.NameServerDaemon(port=0)
        with ns:
            pass
        with pytest.raises(PyroError):
            with ns:
                pass
        ns.close()

    @pytest.mark.network
    def testStartNSfunc(self):
        myIpAddress = Pyro5.socketutil.get_ip_address("", workaround127=True)
        uri1, ns1, bc1 = Pyro5.nameserver.start_ns(host=myIpAddress, port=0, bcport=0, enableBroadcast=False)
        uri2, ns2, bc2 = Pyro5.nameserver.start_ns(host=myIpAddress, port=0, bcport=0, enableBroadcast=True)
        assert isinstance(uri1, Pyro5.core.URI)
        assert isinstance(ns1, Pyro5.nameserver.NameServerDaemon)
        assert bc1 is None
        assert isinstance(bc2, Pyro5.nameserver.BroadcastServer)
        sock = bc2.sock
        assert hasattr(sock, "fileno")
        _ = bc2.processRequest
        ns1.close()
        ns2.close()
        bc2.close()

    def testNSmain(self):
        oldstdout = sys.stdout
        oldstderr = sys.stderr
        try:
            sys.stdout = StringIO()
            sys.stderr = StringIO()
            with pytest.raises(SystemExit):
                Pyro5.nameserver.main(["--invalidarg"])
            assert "usage" in sys.stderr.getvalue()
            sys.stderr.truncate(0)
            sys.stdout.truncate(0)
            with pytest.raises(SystemExit):
                Pyro5.nameserver.main(["-h"])
            assert "show this help message" in sys.stdout.getvalue()
        finally:
            sys.stdout = oldstdout
            sys.stderr = oldstderr

    def testNSCmain(self):
        oldstdout = sys.stdout
        oldstderr = sys.stderr
        try:
            sys.stdout = StringIO()
            sys.stderr = StringIO()
            with pytest.raises(SystemExit):
                Pyro5.nsc.main(["--invalidarg"])
            assert "usage" in sys.stderr.getvalue()
            sys.stderr.truncate(0)
            sys.stdout.truncate(0)
            with pytest.raises(SystemExit):
                Pyro5.nsc.main(["-h"])
            assert "show this help message" in sys.stdout.getvalue()
        finally:
            sys.stdout = oldstdout
            sys.stderr = oldstderr

    def testNSCfunctions(self):
        oldstdout = sys.stdout
        ns = None
        try:
            sys.stdout = StringIO()
            ns = Pyro5.nameserver.NameServer(storageProvider=self.storageProvider)
            with pytest.raises(KeyError):
                Pyro5.nsc.handle_command(ns, "foo", [])
            assert sys.stdout.getvalue().startswith("Error: KeyError ")
            Pyro5.nsc.handle_command(ns, "ping", [])
            assert sys.stdout.getvalue().endswith("ping ok.\n")
            with pytest.raises(NamingError):
                Pyro5.nsc.handle_command(ns, "lookup", ["WeirdName"])
            assert sys.stdout.getvalue().endswith("Error: NamingError - unknown name: WeirdName\n")
            Pyro5.nsc.handle_command(ns, "list", [])
            assert sys.stdout.getvalue().endswith("END LIST \n")
            Pyro5.nsc.handle_command(ns, "listmatching", ["name.$"])
            assert sys.stdout.getvalue().endswith("END LIST - regex 'name.$'\n")
            assert "name1" not in sys.stdout.getvalue()
            Pyro5.nsc.handle_command(ns, "register", ["name1", "PYRO:obj1@hostname:9999"])
            assert sys.stdout.getvalue().endswith("Registered name1\n")
            Pyro5.nsc.handle_command(ns, "remove", ["name2"])
            assert sys.stdout.getvalue().endswith("Nothing removed\n")
            Pyro5.nsc.handle_command(ns, "listmatching", ["name.$"])
            assert "name1 --> PYRO:obj1@hostname:9999" in sys.stdout.getvalue()
            # Pyro5.nsc.handle_command(ns, None, ["removematching", "name.?"])  #  can't be tested, required user input
        finally:
            sys.stdout = oldstdout
            if ns:
                ns.storage.close()

    def testNAT(self):
        uri, ns, bc = Pyro5.nameserver.start_ns(host="", port=0, enableBroadcast=True, nathost="nathosttest", natport=12345)
        assert uri.location == "nathosttest:12345"
        assert ns.uriFor("thing").location == "nathosttest:12345"
        assert bc.nsUri.location != "nathosttest:12345", "broadcast location must not be the NAT location"
        ns.close()
        bc.close()

    def testMetadataRegisterInvalidTypes(self):
        ns = Pyro5.nameserver.NameServer(storageProvider=self.storageProvider)
        with pytest.raises(TypeError):
            ns.register("meta1", "PYRO:meta1@localhost:1111", metadata=12345)   # metadata must be iterable
        with pytest.raises(TypeError):
            ns.register("meta1", "PYRO:meta1@localhost:1111", metadata="string")   # metadata must not be str

    def testMetadataLookupInvalidTypes(self):
        ns = Pyro5.nameserver.NameServer(storageProvider=self.storageProvider)
        with pytest.raises(TypeError):
            ns.yplookup(meta_all=12345)
        with pytest.raises(TypeError):
            ns.yplookup(meta_all="string")
        with pytest.raises(TypeError):
            ns.yplookup(meta_any=12345)
        with pytest.raises(TypeError):
            ns.yplookup(meta_any="string")

    def testMetadata(self):
        self.clearStorage()
        ns = Pyro5.nameserver.NameServer(storageProvider=self.storageProvider)
        # register some names with metadata, and perform simple lookups
        ns.register("meta1", "PYRO:meta1@localhost:1111", metadata={"a", "b", "c"})
        ns.register("meta2", "PYRO:meta2@localhost:2222", metadata={"x", "y", "z"})
        ns.register("meta3", "PYRO:meta3@localhost:3333", metadata=["p", "q", "r", "r", "q"])
        uri = ns.lookup("meta1")
        assert uri.object == "meta1"
        uri, metadata = ns.lookup("meta1", return_metadata=True)
        assert uri.object == "meta1"
        assert set(metadata) == {"a", "b", "c"}
        uri = ns.lookup("meta2")
        assert uri.object == "meta2"
        uri, metadata = ns.lookup("meta2", return_metadata=True)
        assert uri.object == "meta2"
        assert set(metadata) == {"x", "y", "z"}
        uri, metadata = ns.lookup("meta3", return_metadata=True)
        assert uri.object == "meta3"
        assert isinstance(metadata, set)
        assert set(metadata) == {"p", "q", "r"}
        # get a list of everything, without and with metadata
        reg = ns.list()
        assert reg == {'meta1': 'PYRO:meta1@localhost:1111', 'meta2': 'PYRO:meta2@localhost:2222', 'meta3': 'PYRO:meta3@localhost:3333'}
        reg = ns.list(return_metadata=True)
        uri1, meta1 = reg["meta1"]
        uri2, meta2 = reg["meta2"]
        assert uri1 == "PYRO:meta1@localhost:1111"
        assert set(meta1) == {"a", "b", "c"}
        assert uri2 == "PYRO:meta2@localhost:2222"
        assert set(meta2) == {"x", "y", "z"}
        # filter on metadata subset
        reg = ns.yplookup(meta_all={"a", "c"}, return_metadata=False)
        assert len(reg) == 1
        assert reg["meta1"] == "PYRO:meta1@localhost:1111"
        reg = ns.yplookup(meta_all={"a", "c"}, return_metadata=True)
        assert len(reg) == 1
        uri1, meta1 = reg["meta1"]
        assert uri1 == "PYRO:meta1@localhost:1111"
        assert set(meta1) == {"a", "b", "c"}
        reg = ns.yplookup(meta_all={"a", "wrong"})
        assert reg == {}
        reg = ns.yplookup(meta_all={"a", "b", "c", "wrong"})
        assert reg == {}
        reg = ns.yplookup(meta_all={"a", "c", "x"})
        assert reg == {}
        # update some metadata
        with pytest.raises(NamingError):
            ns.set_metadata("notexistingname", set())
        ns.set_metadata("meta1", {"one", "two", "three"})
        uri, meta = ns.lookup("meta1", return_metadata=True)
        assert set(meta) == {"one", "two", "three"}
        # check that a collection is converted to a set
        ns.set_metadata("meta1", ["one", "two", "three", "three", "two"])
        uri, meta = ns.lookup("meta1", return_metadata=True)
        assert isinstance(meta, set)
        assert set(meta) == {"one", "two", "three"}
        # remove record that has some metadata
        ns.remove("meta1")
        ns.remove("meta3")
        assert list(ns.list().keys()) == ["meta2"]
        # other list filters
        reg = ns.list(prefix="meta", return_metadata=True)
        assert len(reg) == 1
        assert set(reg["meta2"][1]) == {"x", "y", "z"}
        reg = ns.list(regex="meta2.*", return_metadata=True)
        assert len(reg) == 1
        assert set(reg["meta2"][1]) == {"x", "y", "z"}
        assert ns.count() == 1

    def testMetadataAny(self):
        self.clearStorage()
        ns = Pyro5.nameserver.NameServer(storageProvider=self.storageProvider)
        # register some names with metadata, and perform simple lookups
        ns.register("meta1", "PYRO:meta1@localhost:1111", metadata={"a", "b", "c"})
        ns.register("meta2", "PYRO:meta2@localhost:2222", metadata={"x", "y", "z"})
        ns.register("meta3", "PYRO:meta3@localhost:2222", metadata={"k", "l", "m"})
        result = ns.yplookup(meta_any={"1", "2", "3"})
        assert result == {}
        result = ns.yplookup(meta_any={"1", "2", "a"})
        assert len(result) == 1
        assert "meta1" in result
        result = ns.yplookup(meta_any={"1", "2", "a", "z"})
        assert len(result) == 2
        assert "meta1" in result
        assert "meta2" in result

    def testEmptyMetadata(self):
        self.clearStorage()
        ns = Pyro5.nameserver.NameServer(storageProvider=self.storageProvider)
        # register some names with metadata, and perform simple lookups
        ns.register("meta1", "PYRO:meta1@localhost:1111", metadata=set())
        uri, meta = ns.lookup("meta1", return_metadata=True)
        assert meta == set()
        registrations = ns.list(return_metadata=True)
        for name in registrations:
            uri, meta = registrations[name]
            assert meta == set()
        ns.set_metadata("meta1", set())

    def testListNoMultipleFilters(self):
        ns = Pyro5.nameserver.NameServer(storageProvider=self.storageProvider)
        with pytest.raises(ValueError):
            ns.list(prefix="a", regex="a")
        with pytest.raises(ValueError):
            ns.yplookup(meta_any={"a"}, meta_all={"a"})


class TestOfflineNameServerTestsSqlStorage(TestOfflineNameServer):
    def setup_method(self):
        super().setup_method()
        self.storageProvider = Pyro5.nameserver.SqlStorage("pyro-test.sqlite")

    def teardown_method(self):
        try:
            super().teardown_method()
        except AttributeError:
            pass
        import glob
        for file in glob.glob("pyro-test.sqlite*"):
            os.remove(file)



import socket
import select
import time
import Pyro5.socketutil
import Pyro5.api


Pyro5.config.SERVERTYPE = "thread"


hostname = socket.gethostname()
my_ip = Pyro5.socketutil.get_ip_address(None, workaround127=True)


@Pyro5.api.expose
class EmbeddedServer(object):
    def multiply(self, x, y):
        return x * y


print("THREADED server type. Initializing services...")
print("Make sure that you don't have a name server running already!\n")
# start a name server with broadcast server
nameserverUri, nameserverDaemon, broadcastServer = Pyro5.nameserver.start_ns(host=my_ip)
assert broadcastServer is not None, "expect a broadcast server to be created"
print("got a Nameserver, uri=%s" % nameserverUri)

# create a Pyro daemon
pyrodaemon = Pyro5.api.Daemon(host=hostname)
serveruri = pyrodaemon.register(EmbeddedServer())
print("server uri=%s" % serveruri)

# register it with the embedded nameserver
nameserverDaemon.nameserver.register("example.eventloop.server", serveruri)

print("")

# Below is our custom event loop.
# Because this particular server runs the different daemons using the "tread" server type,
# there is no built in way of combining the different event loops and server sockets.
# We have to write our own multiplexing server event loop, and dispatch the requests
# to the server that they belong to.
# It is a bit silly to do it this way because the choice for a threaded server type
# has already been made-- so you could just as well run the different daemons' request loops
# each in their own thread and avoid writing this integrated event loop altogether.
# But for the sake of example we write out our own loop:

while True:
    print(time.asctime(), "Waiting for requests...")
    # create sets of the socket objects we will be waiting on
    # (a set provides fast lookup compared to a list)
    nameserverSockets = set(nameserverDaemon.sockets)
    pyroSockets = set(pyrodaemon.sockets)
    rs = [broadcastServer]  # only the broadcast server is directly usable as a select() object
    rs.extend(nameserverSockets)
    rs.extend(pyroSockets)
    rs, _, _ = select.select(rs, [], [], 3)
    eventsForNameserver = []
    eventsForDaemon = []
    for s in rs:
        if s is broadcastServer:
            print("Broadcast server received a request")
            broadcastServer.processRequest()
        elif s in nameserverSockets:
            eventsForNameserver.append(s)
        elif s in pyroSockets:
            eventsForDaemon.append(s)
    if eventsForNameserver:
        print("Nameserver received a request")
        nameserverDaemon.events(eventsForNameserver)
    if eventsForDaemon:
        print("Daemon received a request")
        pyrodaemon.events(eventsForDaemon)



import Pyro5.api
import Pyro5.socketutil
import bench


Pyro5.api.serve({
        bench.bench: "example.benchmark"
    },
    daemon=Pyro5.api.Daemon(host=Pyro5.socketutil.get_ip_address("", True)),
    use_ns=False)



from Pyro5.api import expose, serve, config, register_class_to_dict, register_dict_to_class
import mycustomclasses


# use serpent
config.SERIALIZER = "serpent"


# register the special serialization hooks
def thingy_class_to_dict(obj):
    print("{serializer hook, converting to dict: %s}" % obj)
    return {
        "__class__": "waheeee-custom-thingy",
        "number-attribute": obj.number
    }


def thingy_dict_to_class(classname, d):
    print("{deserializer hook, converting to class: %s}" % d)
    return mycustomclasses.Thingy(d["number-attribute"])


def otherthingy_dict_to_class(classname, d):
    print("{deserializer hook, converting to class: %s}" % d)
    return mycustomclasses.OtherThingy(d["number"])


# for 'Thingy' we register both serialization and deserialization hooks
register_dict_to_class("waheeee-custom-thingy", thingy_dict_to_class)
register_class_to_dict(mycustomclasses.Thingy, thingy_class_to_dict)

# for 'OtherThingy' we only register a deserialization hook (and for serialization depend on serpent's default behavior)
register_dict_to_class("mycustomclasses.OtherThingy", otherthingy_dict_to_class)


# regular Pyro server stuff

@expose
class Server(object):
    def method(self, arg):
        print("\nmethod called, arg=", arg)
        response = mycustomclasses.Thingy(999)
        return response

    def othermethod(self, arg):
        print("\nothermethod called, arg=", arg)
        response = mycustomclasses.OtherThingy(999)
        return response


serve(
    {
        Server: "example.customclasses"
    },
    use_ns=False)



from Pyro5.api import Proxy
import Pyro5.errors


uri = input("Enter the URI of the thirdparty library object: ").strip()
with Proxy(uri) as remote:
    print(remote.method("how are you?"))

    try:
        print(remote.weird())
    except Pyro5.errors.SerializeError:
        print("couldn't call weird() due to serialization error of the result value! (is ok)")

    try:
        print(remote.private())    # we can call this if full class is exposed...
    except AttributeError:
        print("couldn't call private(), it doesn't seem to be exposed! (is ok)")



import time
from Pyro5.api import Proxy


print("First make sure one of the gui servers is running.")
print("Enter the object uri that was printed:")
uri = input().strip()
guiserver = Proxy(uri)

guiserver.message("Hello there!")
time.sleep(0.5)
guiserver.message("How's it going?")
time.sleep(2)

for i in range(20):
    guiserver.message("Counting {0}".format(i))

guiserver.message("now calling the sleep method with 5 seconds")
guiserver.sleep(5)
print("done!")



import time
import threading
from Pyro5.api import Proxy


serv = Proxy("PYRONAME:example.servertypes")

print("--------------------------------------------------------------")
print("    This part is independent of the type of the server.       ")
print("--------------------------------------------------------------")
print("Calling 5 times oneway method. Should return immediately.")
serv.reset()
begin = time.time()
serv.onewaydelay()
serv.onewaydelay()
serv.onewaydelay()
serv.onewaydelay()
serv.onewaydelay()
print("Done with the oneway calls.")
completed = serv.getcount()
print("Number of completed calls in the server: %d" % completed)
print("  (this should be 0, because all 5 calls are still busy in the background)")
if completed > 0:
    raise SystemExit("error: oneway calls should run in the background!")
print()
print("Calling normal delay 5 times. They will all be processed by the same server thread because we're using the same proxy.")
r = serv.delay()
print("  call processed by: %s" % r)
r = serv.delay()
print("  call processed by: %s" % r)
r = serv.delay()
print("  call processed by: %s" % r)
r = serv.delay()
print("  call processed by: %s" % r)
r = serv.delay()
print("  call processed by: %s" % r)
time.sleep(2)
print("Number of completed calls in the server: %d" % serv.getcount())
print("  (this should be 10, because by now the 5 oneway calls have completed as well)")
serv.reset()

print("\n--------------------------------------------------------------")
print("    This part depends on the type of the server.              ")
print("--------------------------------------------------------------")
print("Creating 5 threads that each call the server at the same time.")
serverconfig = serv.getconfig()
if serverconfig["SERVERTYPE"] == "thread":
    print("Servertype is thread. All calls will run in parallel.")
    print("The time this will take is 1 second (every thread takes 1 second in parallel).")
    print("You will see that the requests are handled by different server threads.")
elif serverconfig["SERVERTYPE"] == "multiplex":
    print("Servertype is multiplex. The calls will need to get in line.")
    print("The time this will take is 5 seconds (every thread takes 1 second sequentially).")
    print("You will see that the requests are handled by a single server thread.")
else:
    print("Unknown servertype")


def func(uri):
    # This will run in a thread. Create a proxy just for this thread:
    with Proxy(uri) as p:
        processed = p.delay()
        print("[ thread %s called delay, processed by: %s ]  " % (threading.current_thread().name, processed))


serv._pyroBind()  # simplify the uri
threads = []
for i in range(5):
    t = threading.Thread(target=func, args=[serv._pyroUri])
    t.daemon = True
    threads.append(t)
    t.start()
print("Waiting for threads to finish:")
for t in threads:
    t.join()
print("Done. Number of completed calls in the server: %d" % serv.getcount())



from Pyro5.api import Proxy
from diffiehellman import DiffieHellman


dh = DiffieHellman(group=14)

with Proxy("PYRONAME:example.dh.keyexchange") as keyex:
    print("exchange public keys...")
    other_key = keyex.exchange_key(dh.public_key)
    print("got server public key, creating shared secret key...")
    dh.make_shared_secret_and_key(other_key)
    print("shared secret key = ", dh.key.hex())
    print("(check the server output to see the same shared private key)")



import socket
import pytest
from Pyro5.compatibility import Pyro4


def test_compat_config():
    import Pyro4
    conf = Pyro4.config.asDict()
    assert conf["NS_PORT"] == 9090
    Pyro4.config.NS_PORT = 12345
    conf = Pyro4.config.asDict()
    assert conf["NS_PORT"] == 12345
    Pyro4.config.NS_PORT = 9090


def test_compat_layer():
    from Pyro4 import naming
    from Pyro4 import socketutil
    from Pyro4 import util
    try:
        _ = 1//0
    except ZeroDivisionError:
        tb = util.getPyroTraceback()
        assert len(tb) == 3
        assert "Traceback" in tb[0]
        assert "zero" in tb[2]
    assert 4 == socketutil.getIpVersion("127.0.0.1")
    assert 6 == socketutil.getIpVersion("::1")
    Pyro4.URI("PYRO:test@localhost:5555")
    p = Pyro4.Proxy("PYRO:test@localhost:5555")
    Pyro4.BatchProxy(p)
    Pyro4.Daemon()
    assert socketutil.getIpAddress("localhost", ipVersion=4).startswith("127.0")
    if socket.has_ipv6:
        try:
            assert ":" in socketutil.getIpAddress("localhost", ipVersion=6)
        except socket.error as x:
            if str(x) != "unable to determine IPV6 address":
                raise
    assert "127.0.0.1" == socketutil.getIpAddress("127.0.0.1")
    assert "::1" == socketutil.getIpAddress("::1")
    assert "127.0.0.1" == socketutil.getInterfaceAddress("127.0.0.1")
    with pytest.raises(NotImplementedError):
        naming.NameServer()
    with pytest.raises(NotImplementedError):
        _ = p._pyroHmacKey
    with pytest.raises(NotImplementedError):
        p._pyroHmacKey = b"fail"




from Pyro5.api import expose, oneway, serve
import Pyro5.errors


class CallbackServer(object):
    @expose
    @oneway
    def doCallback(self, callback):
        print("\n\nserver: doing callback 1 to client")
        callback._pyroClaimOwnership()
        try:
            callback.call1()
        except Exception:
            print("got an exception from the callback:")
            print("".join(Pyro5.errors.get_pyro_traceback()))
        print("\n\nserver: doing callback 2 to client")
        try:
            callback.call2()
        except Exception:
            print("got an exception from the callback:")
            print("".join(Pyro5.errors.get_pyro_traceback()))
        print("server: callbacks done.\n")


serve({
    CallbackServer: "example.callback2"
})



"""
Tests for the core logic.

Pyro - Python Remote Objects.  Copyright by Irmen de Jong (irmen@razorvine.net).
"""

import os
import copy
import uuid
import pytest
import importlib
import logging
import Pyro5.core
import Pyro5.callcontext
import Pyro5.client
import Pyro5.errors
import Pyro5.configure
import Pyro5.server
from Pyro5 import config


class TestCore:
    def test_uri(self):
        with pytest.raises(Pyro5.errors.PyroError):
            Pyro5.core.URI("burp")
        u1 = Pyro5.core.URI("PYRO:obj@host:5555")
        u2 = copy.copy(u1)
        assert str(u1) == str(u2)
        assert u1 == u2
        assert u1 is not u2

    def test_unix_uri(self):
        p = Pyro5.core.URI("PYRO:12345@./u:/tmp/sockname")
        assert p.object == "12345"
        assert p.sockname == "/tmp/sockname"
        p = Pyro5.core.URI("PYRO:12345@./u:../sockname")
        assert p.object == "12345"
        assert p.sockname == "../sockname"
        p = Pyro5.core.URI("PYRO:12345@./u:/path with spaces/sockname  ")
        assert p.object == "12345"
        assert p.sockname == "/path with spaces/sockname  "

    def testConfig(self):
        assert type(config.COMPRESSION) is bool
        assert type(config.NS_PORT) is int
        cfgdict = config.as_dict()
        assert type(cfgdict) is dict
        assert "COMPRESSION" in cfgdict
        assert config.COMPRESSION == cfgdict["COMPRESSION"]

    def testConfigDefaults(self):
        # some security sensitive settings:
        config.reset(False)   # reset the config to default
        assert config.HOST == "localhost"
        assert config.NS_HOST == "localhost"
        assert config.SERIALIZER == "serpent"

    def testConfigValid(self):
        with pytest.raises(AttributeError):
            config.XYZ_FOOBAR = True  # don't want to allow weird config names

    def testConfigParseBool(self):
        config = Pyro5.configure.Configuration()
        assert type(config.COMPRESSION) is bool
        os.environ["PYRO_COMPRESSION"] = "yes"
        config.reset()
        assert config.COMPRESSION
        os.environ["PYRO_COMPRESSION"] = "off"
        config.reset()
        assert not config.COMPRESSION
        os.environ["PYRO_COMPRESSION"] = "foobar"
        with pytest.raises(ValueError):
            config.reset()
        del os.environ["PYRO_COMPRESSION"]
        config.reset()

    def testConfigDump(self):
        config = Pyro5.configure.Configuration()
        dump = config.dump()
        assert "version:" in dump
        assert "LOGLEVEL" in dump

    def testLogInit(self):
        _ = logging.getLogger("Pyro5")
        os.environ["PYRO_LOGLEVEL"] = "DEBUG"
        os.environ["PYRO_LOGFILE"] = "{stderr}"
        importlib.reload(Pyro5)
        _ = logging.getLogger("Pyro5")
        os.environ["PYRO_LOGFILE"] = "Pyro.log"
        importlib.reload(Pyro5)
        _ = logging.getLogger("Pyro5")
        del os.environ["PYRO_LOGLEVEL"]
        del os.environ["PYRO_LOGFILE"]
        importlib.reload(Pyro5)
        _ = logging.getLogger("Pyro5")

    def testUriStrAndRepr(self):
        uri = "PYRONAME:some_obj_name"
        p = Pyro5.core.URI(uri)
        assert str(p) == uri
        uri = "PYRONAME:some_obj_name@host.com"
        p = Pyro5.core.URI(uri)
        assert str(p) == uri + ":" + str(config.NS_PORT)  # a PYRONAME uri with a hostname gets a port too if omitted
        uri = "PYRONAME:some_obj_name@host.com:8888"
        p = Pyro5.core.URI(uri)
        assert str(p) == uri
        expected = "<Pyro5.core.URI at 0x%x; PYRONAME:some_obj_name@host.com:8888>" % id(p)
        assert repr(p) == expected
        uri = "PYRO:12345@host.com:9999"
        p = Pyro5.core.URI(uri)
        assert str(p) == uri
        uri = "PYRO:12345@./u:sockname"
        p = Pyro5.core.URI(uri)
        assert str(p) == uri
        uri = "PYRO:12345@./u:sockname"
        assert str(p) == uri
        assert type(p.sockname) is str
        uri = "PYRO:12345@./u:sock name with strings"
        p = Pyro5.core.URI(uri)
        assert str(p) == uri

    def testUriParsingPyro(self):
        p = Pyro5.core.URI("PYRONAME:some_obj_name")
        assert p.protocol == "PYRONAME"
        assert p.object == "some_obj_name"
        assert p.host is None
        assert p.sockname is None
        assert p.port is None
        p = Pyro5.core.URI("PYRONAME:some_obj_name@host.com:9999")
        assert p.protocol == "PYRONAME"
        assert p.object == "some_obj_name"
        assert p.host == "host.com"
        assert p.port == 9999

        p = Pyro5.core.URI("PYRO:12345@host.com:4444")
        assert p.protocol == "PYRO"
        assert p.object == "12345"
        assert p.host == "host.com"
        assert p.sockname is None
        assert p.port == 4444
        assert p.location == "host.com:4444"
        p = Pyro5.core.URI("PYRO:12345@./u:sockname")
        assert p.object == "12345"
        assert p.sockname == "sockname"
        p = Pyro5.core.URI("PYRO:12345@./u:/tmp/sockname")
        assert p.object == "12345"
        assert p.sockname == "/tmp/sockname"
        p = Pyro5.core.URI("PYRO:12345@./u:/path with spaces/sockname  ")
        assert p.object == "12345"
        assert p.sockname == "/path with spaces/sockname  "
        p = Pyro5.core.URI("PYRO:12345@./u:../sockname")
        assert p.object == "12345"
        assert p.sockname == "../sockname"
        p = Pyro5.core.URI("pyro:12345@host.com:4444")
        assert p.protocol == "PYRO"
        assert p.object == "12345"
        assert p.host == "host.com"
        assert p.sockname is None
        assert p.port == 4444

    def testUriParsingIpv6(self):
        p = Pyro5.core.URI("pyro:12345@[::1]:4444")
        assert p.host == "::1"
        assert p.location == "[::1]:4444"
        with pytest.raises(Pyro5.errors.PyroError) as e:
            Pyro5.core.URI("pyro:12345@[[::1]]:4444")
        assert str(e.value) == "invalid ipv6 address: enclosed in too many brackets"
        with pytest.raises(Pyro5.errors.PyroError) as e:
            Pyro5.core.URI("pyro:12345@[must_be_numeric_here]:4444")
        assert str(e.value) == "invalid ipv6 address: the part between brackets must be a numeric ipv6 address"

    def testUriParsingPyroname(self):
        p = Pyro5.core.URI("PYRONAME:objectname")
        assert p.protocol == "PYRONAME"
        assert p.object == "objectname"
        assert p.host is None
        assert p.port is None
        p = Pyro5.core.URI("PYRONAME:objectname@nameserverhost")
        assert p.protocol == "PYRONAME"
        assert p.object == "objectname"
        assert p.host == "nameserverhost"
        assert p.port == config.NS_PORT  # Pyroname uri with host gets a port too if not specified
        p = Pyro5.core.URI("PYRONAME:objectname@nameserverhost:4444")
        assert p.protocol == "PYRONAME"
        assert p.object == "objectname"
        assert p.host == "nameserverhost"
        assert p.port == 4444
        p = Pyro5.core.URI("PyroName:some_obj_name@host.com:9999")
        assert p.protocol == "PYRONAME"
        p = Pyro5.core.URI("pyroname:some_obj_name@host.com:9999")
        assert p.protocol == "PYRONAME"

    def testUriParsingPyrometa(self):
        p = Pyro5.core.URI("PYROMETA:meta")
        assert p.protocol == "PYROMETA"
        assert p.object == {"meta"}
        assert p.host is None
        assert p.port is None
        p = Pyro5.core.URI("PYROMETA:meta1,meta2,meta2@nameserverhost")
        assert p.protocol == "PYROMETA"
        assert p.object == {"meta1", "meta2"}
        assert p.host == "nameserverhost"
        assert p.port == config.NS_PORT  # PyroMeta uri with host gets a port too if not specified
        p = Pyro5.core.URI("PYROMETA:meta@nameserverhost:4444")
        assert p.protocol == "PYROMETA"
        assert p.object == {"meta"}
        assert p.host == "nameserverhost"
        assert p.port == 4444
        p = Pyro5.core.URI("PyroMeta:meta1,meta2@host.com:9999")
        assert p.protocol == "PYROMETA"
        p = Pyro5.core.URI("PyroMeta:meta1,meta2@host.com:9999")
        assert p.protocol == "PYROMETA"

    def testInvalidUris(self):
        with pytest.raises(TypeError):
            Pyro5.core.URI(None)
        with pytest.raises(TypeError):
            Pyro5.core.URI(99999)
        with pytest.raises(Pyro5.errors.PyroError):
            Pyro5.core.URI("")
        with pytest.raises(Pyro5.errors.PyroError):
            Pyro5.core.URI(" ")
        with pytest.raises(Pyro5.errors.PyroError):
            Pyro5.core.URI("a")
        with pytest.raises(Pyro5.errors.PyroError):
            Pyro5.core.URI("PYR")
        with pytest.raises(Pyro5.errors.PyroError):
            Pyro5.core.URI("PYRO")
        with pytest.raises(Pyro5.errors.PyroError):
            Pyro5.core.URI("PYRO:")
        with pytest.raises(Pyro5.errors.PyroError):
            Pyro5.core.URI("PYRO::")
        with pytest.raises(Pyro5.errors.PyroError):
            Pyro5.core.URI("PYRO:a")
        with pytest.raises(Pyro5.errors.PyroError):
            Pyro5.core.URI("PYRO:x@")
        with pytest.raises(Pyro5.errors.PyroError):
            Pyro5.core.URI("PYRO:x@hostname")
        with pytest.raises(Pyro5.errors.PyroError):
            Pyro5.core.URI("PYRO:@hostname:portstr")
        with pytest.raises(Pyro5.errors.PyroError):
            Pyro5.core.URI("PYRO:@hostname:7766")
        with pytest.raises(Pyro5.errors.PyroError):
            Pyro5.core.URI("PYRO:objid@hostname:7766:bogus")
        with pytest.raises(Pyro5.errors.PyroError):
            Pyro5.core.URI("PYRO:obj id@hostname:7766")
        with pytest.raises(Pyro5.errors.PyroError):
            Pyro5.core.URI("PYROLOC:objname")
        with pytest.raises(Pyro5.errors.PyroError):
            Pyro5.core.URI("PYROLOC:objname@host")
        with pytest.raises(Pyro5.errors.PyroError):
            Pyro5.core.URI("PYROLOC:objectname@hostname:4444")
        with pytest.raises(Pyro5.errors.PyroError):
            Pyro5.core.URI("PYRONAME:")
        with pytest.raises(Pyro5.errors.PyroError):
            Pyro5.core.URI("PYRONAME:obj name@nameserver:bogus")
        with pytest.raises(Pyro5.errors.PyroError):
            Pyro5.core.URI("PYRONAME:objname@nameserver:bogus")
        with pytest.raises(Pyro5.errors.PyroError):
            Pyro5.core.URI("PYRONAME:objname@nameserver:7766:bogus")
        with pytest.raises(Pyro5.errors.PyroError):
            Pyro5.core.URI("PYROMETA:")
        with pytest.raises(Pyro5.errors.PyroError):
            Pyro5.core.URI("PYROMETA:meta@nameserver:bogus")
        with pytest.raises(Pyro5.errors.PyroError):
            Pyro5.core.URI("PYROMETA:meta@nameserver:7766:bogus")
        with pytest.raises(Pyro5.errors.PyroError):
            Pyro5.core.URI("PYROMETA:meta1, m2 ,m3@nameserver:7766:bogus")
        with pytest.raises(Pyro5.errors.PyroError):
            Pyro5.core.URI("FOOBAR:")
        with pytest.raises(Pyro5.errors.PyroError):
            Pyro5.core.URI("FOOBAR:objid@hostname:7766")
        with pytest.raises(Pyro5.errors.PyroError):
            Pyro5.core.URI("PYRO:12345@./u:sockname:9999")

    def testUriUnicode(self):
        p = Pyro5.core.URI("PYRO:12345@host.com:4444")
        assert p.protocol == "PYRO"
        assert p.object == "12345"
        assert p.host == "host.com"
        assert type(p.protocol) is str
        assert type(p.object) is str
        assert type(p.host) is str
        assert p.sockname is None
        assert p.port == 4444

        uri = "PYRO:12345@hostname:9999"
        p = Pyro5.core.URI(uri)
        unicodeuri = "PYRO:weirdchars" + chr(0x20ac) + "@host" + chr(0x20AC) + ".com:4444"
        pu = Pyro5.core.URI(unicodeuri)
        assert pu.protocol == "PYRO"
        assert pu.host == "host" + chr(0x20AC) + ".com"
        assert pu.object == "weirdchars" + chr(0x20AC)
        assert str(pu) == "PYRO:weirdchars" + chr(0x20ac) + "@host" + chr(0x20ac) + ".com:4444"
        expected = ("<Pyro5.core.URI at 0x%x; PYRO:weirdchars" + chr(0x20ac) + "@host" + chr(0x20ac) + ".com:4444>") % id(pu)
        assert repr(pu) == expected
        assert str(pu) == "PYRO:weirdchars" + chr(0x20ac) + "@host" + chr(0x20ac) + ".com:4444"

    def testUriCopy(self):
        p1 = Pyro5.core.URI("PYRO:12345@hostname:9999")
        p2 = Pyro5.core.URI(p1)
        p3 = copy.copy(p1)
        assert p2.protocol == p1.protocol
        assert p2.host == p1.host
        assert p2.port == p1.port
        assert p2.object == p1.object
        assert p2 == p1
        assert p3.protocol == p1.protocol
        assert p3.host == p1.host
        assert p3.port == p1.port
        assert p3.object == p1.object
        assert p3 == p1

    def testUriSubclassCopy(self):
        class SubURI(Pyro5.core.URI):
            pass
        u = SubURI("PYRO:12345@hostname:9999")
        u2 = copy.copy(u)
        assert isinstance(u2, SubURI)

    def testUriEqual(self):
        p1 = Pyro5.core.URI("PYRO:12345@host.com:9999")
        p2 = Pyro5.core.URI("PYRO:12345@host.com:9999")
        p3 = Pyro5.core.URI("PYRO:99999@host.com:4444")
        assert p2 == p1
        assert p1 != p3
        assert p2 != p3
        assert p1 == p2
        assert p1 != p3
        assert p2 != p3
        assert p1 == p2
        assert p1 != p3
        assert p2 != p3
        assert hash(p1) == hash(p2)
        assert hash(p1) != hash(p3)
        p2.port = 4444
        p2.object = "99999"
        assert p1 != p2
        assert p3 == p2
        assert p1 != p2
        assert p2 == p3
        assert p1 != p2
        assert p2 == p3
        assert hash(p1) != hash(p2)
        assert hash(p2) == hash(p3)
        assert p1 != 42

    def testLocation(self):
        assert Pyro5.core.URI.isUnixsockLocation("./u:name")
        assert not(Pyro5.core.URI.isUnixsockLocation("./p:name"))
        assert not(Pyro5.core.URI.isUnixsockLocation("./x:name"))
        assert not(Pyro5.core.URI.isUnixsockLocation("foobar"))

    def testCallContext(self):
        ctx = Pyro5.callcontext.current_context
        corr_id = uuid.UUID('1897022f-c481-4117-a4cc-cbd1ca100582')
        ctx.correlation_id = corr_id
        d = ctx.to_global()
        assert isinstance(d, dict)
        assert d["correlation_id"] == corr_id
        corr_id2 = uuid.UUID('67b05ad9-2d6a-4ed8-8ed5-95cba68b4cf9')
        d["correlation_id"] = corr_id2
        ctx.from_global(d)
        assert Pyro5.callcontext.current_context.correlation_id == corr_id2
        Pyro5.callcontext.current_context.correlation_id = None




import time
import threading
import Pyro5.api
import Pyro5.errors


proxy = Pyro5.api.locate_ns()    # grab a proxy for the name server


print("Main thread:", threading.current_thread())
proxy.ping()    # call it, the proxy is now connected and bound to the main thread


# trying to use the proxy in a different thread is not possible,
# and Pyro will raise an exception to tell you that:

def other_thread_call():
    try:
        proxy.ping()
        print("You should not see this!! the call succeeded in thread", threading.current_thread())
    except Pyro5.errors.PyroError as x:
        print("Expected exception in thread", threading.current_thread())
        print("Exception was: ", x)


print()
threading.Thread(target=other_thread_call).start()
time.sleep(1)


# SOLUTION 1:  create a new proxy in the other thread.

def new_proxy_thread_call(uri):
    proxy = Pyro5.api.Proxy(uri)
    proxy.ping()
    print("Solution 1. The call succeeded in thread", threading.current_thread())


print()
threading.Thread(target=new_proxy_thread_call, args=(proxy._pyroUri,)).start()
time.sleep(1)


# SOLUTION 2:  transfer ownership of our proxy to the other thread.

def new_owner_thread_call(proxy):
    proxy._pyroClaimOwnership()
    proxy.ping()
    print("Solution 2. The call succeeded in thread", threading.current_thread())


print()
threading.Thread(target=new_owner_thread_call, args=(proxy,)).start()
time.sleep(1)


# however, we are no longer the owner of the proxy now, so any new calls will fail for us
print()
try:
    proxy.ping()
    print("You should not see this!! the call succeeded in thread", threading.current_thread())
except Pyro5.errors.PyroError as x:
    print("Expected exception in thread", threading.current_thread())
    print("Exception was: ", x)



from Pyro5.api import Daemon, serve
import excep


def my_error_handler(daemon, client_sock, method, vargs, kwargs, exception):
    print("\nERROR IN METHOD CALL USER CODE:")
    print(" client={} method={} exception={}".format(client_sock, method.__qualname__, repr(exception)))


daemon = Daemon()
daemon.methodcall_error_handler = my_error_handler

serve(
    {
        excep.TestClass: "example.exceptions"
    },
    daemon=daemon,
    use_ns=True, verbose=True)



import Pyro5.errors
import Pyro5.api


huge_object = [42] * 10000
simple_object = {"message": "hello", "irrelevant": huge_object}

print("First start the built-in test echo server with something like:")
print("$ python -m Pyro5.utils.echoserver")
print("Enter the server's uri that was printed:")
uri = input().strip()
echoserver = Pyro5.api.Proxy(uri)

Pyro5.config.MAX_MESSAGE_SIZE = 2**32
print("\nSending big data with virtually no limit on message size...")
response = echoserver.echo(simple_object)
print("success.")

try:
    Pyro5.config.MAX_MESSAGE_SIZE = 2500
    print("\nSending big data with a limit on message size...")
    response = echoserver.echo(simple_object)
    print("Hmm, this should have raised an exception")
except Pyro5.errors.MessageTooLargeError as x:
    print("EXCEPTION (expected):", x)



from Pyro5.api import Proxy


uri = input("enter the server uri: ").strip()
if "\\x00" in uri:
    uri=uri.replace("\\x00", "\x00")
    print("(uri contains 0-byte)")

with Proxy(uri) as p:
    response = p.message("Hello there!")
    print("Response was:", response)



import copy
import pytest
import time
import Pyro5.client
import Pyro5.server
import Pyro5.errors
from Pyro5 import config


class TestProxy:
    def testBasics(self):
        with pytest.raises(Pyro5.errors.PyroError):
            Pyro5.client.Proxy("burp")
        p1 = Pyro5.client.Proxy("PYRO:obj@host:5555")
        p1._pyroHandshake = "milkshake"
        p1._pyroTimeout = 42
        p1._pyroSeq = 100
        p1._pyroMaxRetries = 99
        p1._pyroRawWireResponse = True
        p2 = copy.copy(p1)
        assert p1 == p2
        assert p1 is not p2
        assert p1._pyroUri == p2._pyroUri
        assert p1._pyroHandshake == p2._pyroHandshake
        assert p1._pyroTimeout == p2._pyroTimeout
        assert p1._pyroMaxRetries == p2._pyroMaxRetries
        assert p1._pyroRawWireResponse == p2._pyroRawWireResponse
        assert p2._pyroSeq == 0

    def testProxyCopy(self):
        u = Pyro5.core.URI("PYRO:12345@hostname:9999")
        p1 = Pyro5.client.Proxy(u)
        p2 = copy.copy(p1)  # check that most basic copy also works
        assert p2 == p1
        assert p2._pyroOneway == set()
        p1._pyroAttrs = set("abc")
        p1._pyroTimeout = 42
        p1._pyroOneway = set("def")
        p1._pyroMethods = set("ghi")
        p1._pyroHandshake = "apples"
        p2 = copy.copy(p1)
        assert p2 == p1
        assert p2._pyroUri == p1._pyroUri
        assert p2._pyroOneway == p1._pyroOneway
        assert p2._pyroMethods == p1._pyroMethods
        assert p2._pyroAttrs == p1._pyroAttrs
        assert p2._pyroTimeout == p1._pyroTimeout
        assert p2._pyroHandshake == p1._pyroHandshake
        p1._pyroRelease()
        p2._pyroRelease()

    def testProxySubclassCopy(self):
        class ProxySub(Pyro5.client.Proxy):
            pass
        p = ProxySub("PYRO:12345@hostname:9999")
        p2 = copy.copy(p)
        assert isinstance(p2, ProxySub)
        p._pyroRelease()
        p2._pyroRelease()

    def testBatchProxyAdapterCopy(self):
        with Pyro5.client.Proxy("PYRO:12345@hostname:9999") as proxy:
            batchproxy = Pyro5.client.BatchProxy(proxy)
            p2 = copy.copy(batchproxy)
            assert isinstance(p2, Pyro5.client.BatchProxy)

    def testProxyOffline(self):
        # only offline stuff here.
        # online stuff needs a running daemon, so we do that in another test, to keep this one simple
        with pytest.raises(Pyro5.errors.PyroError):
            Pyro5.client.Proxy("999")
        p1 = Pyro5.client.Proxy("PYRO:9999@localhost:15555")
        p2 = Pyro5.client.Proxy(Pyro5.core.URI("PYRO:9999@localhost:15555"))
        assert p2._pyroUri == p1._pyroUri
        assert p1._pyroConnection is None
        p1._pyroRelease()
        p1._pyroRelease()
        # try copying a not-connected proxy
        p3 = copy.copy(p1)
        assert p3._pyroConnection is None
        assert p1._pyroConnection is None
        assert p1._pyroUri == p3._pyroUri
        assert p1._pyroUri is not p3._pyroUri
        p3._pyroRelease()

    def testProxySerializerOverride(self):
        serializer = config.SERIALIZER
        try:
            with pytest.raises(ValueError) as x:
                config.SERIALIZER = "~invalid~"
                Pyro5.client.Proxy("PYRO:obj@localhost:5555")
            assert "unknown" in str(x.value)
        finally:
            config.SERIALIZER = serializer
        try:
            with pytest.raises(KeyError) as x:
                proxy = Pyro5.client.Proxy("PYRO:obj@localhost:5555")
                proxy._pyroSerializer = "~invalidoverride~"
                proxy._pyroConnection = "FAKE"
                proxy.methodcall()
            assert "invalidoverride" in str(x.value)
        finally:
            proxy._pyroConnection = None
            config.SERIALIZER = serializer

    def testProxyDirMetadata(self):
        with Pyro5.client.Proxy("PYRO:9999@localhost:15555") as p:
            # metadata isn't loaded
            assert '__hash__' in dir(p)
            assert "ping" not in dir(p)
            # emulate obtaining metadata
            p._pyroAttrs = {"prop"}
            p._pyroMethods = {"ping"}
            assert "__hash__" in dir(p)
            assert "prop" in dir(p)
            assert "ping" in dir(p)

    def testProxySettings(self):
        p1 = Pyro5.client.Proxy("PYRO:9999@localhost:15555")
        p2 = Pyro5.client.Proxy("PYRO:9999@localhost:15555")
        p1._pyroOneway.add("method")
        p1._pyroAttrs.add("attr")
        p1._pyroMethods.add("method2")
        assert "method" in p1._pyroOneway
        assert "attr" in p1._pyroAttrs
        assert "method2" in  p1._pyroMethods
        assert "method" not in p2._pyroOneway
        assert "attr" not in p2._pyroAttrs
        assert "method2" not in p2._pyroMethods
        assert p1._pyroOneway is not p2._pyroOneway, "p1 and p2 should have different oneway tables"
        assert p1._pyroAttrs is not p2._pyroAttrs, "p1 and p2 should have different attr tables"
        assert p1._pyroMethods is not p2._pyroMethods, "p1 and p2 should have different method tables"
        p1._pyroRelease()
        p2._pyroRelease()

    def testProxyWithStmt(self):
        class ConnectionMock(object):
            closeCalled = False
            keep_open = False

            def close(self):
                self.closeCalled = True

        connMock = ConnectionMock()
        # first without a 'with' statement
        p = Pyro5.client.Proxy("PYRO:9999@localhost:15555")
        p._pyroConnection = connMock
        assert not(connMock.closeCalled)
        p._pyroRelease()
        assert p._pyroConnection is None
        assert connMock.closeCalled

        connMock = ConnectionMock()
        with Pyro5.client.Proxy("PYRO:9999@localhost:15555") as p:
            p._pyroConnection = connMock
        assert p._pyroConnection is None
        assert connMock.closeCalled
        connMock = ConnectionMock()
        with pytest.raises(ZeroDivisionError):
            with Pyro5.client.Proxy("PYRO:9999@localhost:15555") as p:
                p._pyroConnection = connMock
                print(1 // 0)  # cause an error
        assert p._pyroConnection is None
        assert connMock.closeCalled
        p = Pyro5.client.Proxy("PYRO:9999@localhost:15555")
        with p:
            assert p._pyroUri
        p._pyroRelease()

    def testNoConnect(self):
        wrongUri = Pyro5.core.URI("PYRO:foobar@localhost:59999")
        with pytest.raises(Pyro5.errors.CommunicationError):
            with Pyro5.client.Proxy(wrongUri) as p:
                p.ping()

    def testTimeoutGetSet(self):
        class ConnectionMock(object):
            def __init__(self):
                self.timeout = config.COMMTIMEOUT
                self.keep_open = False

            def close(self):
                pass

        config.COMMTIMEOUT = None
        p = Pyro5.client.Proxy("PYRO:obj@host:555")
        assert p._pyroTimeout is None
        p._pyroTimeout = 5
        assert p._pyroTimeout == 5
        p = Pyro5.client.Proxy("PYRO:obj@host:555")
        p._pyroConnection = ConnectionMock()
        assert p._pyroTimeout is None
        p._pyroTimeout = 5
        assert p._pyroTimeout == 5
        assert p._pyroConnection.timeout == 5
        config.COMMTIMEOUT = 2
        p = Pyro5.client.Proxy("PYRO:obj@host:555")
        p._pyroConnection = ConnectionMock()
        assert p._pyroTimeout == 2
        assert p._pyroConnection.timeout == 2
        p._pyroTimeout = None
        assert p._pyroTimeout is None
        assert p._pyroConnection.timeout is None
        config.COMMTIMEOUT = None
        p._pyroRelease()

    def testCallbackDecorator(self):
        # just test the decorator itself, testing the callback
        # exception handling is kinda hard in unit tests. Maybe later.
        class Test(object):
            @Pyro5.server.callback
            def method(self):
                pass

            def method2(self):
                pass

        t = Test()
        assert getattr(t.method, "_pyroCallback", False)
        assert not getattr(t.method2, "_pyroCallback", False)

    def testProxyEquality(self):
        p1 = Pyro5.client.Proxy("PYRO:thing@localhost:15555")
        p2 = Pyro5.client.Proxy("PYRO:thing@localhost:15555")
        p3 = Pyro5.client.Proxy("PYRO:other@machine:16666")
        assert p1 == p2
        assert not(p1 != p2)
        assert not(p1 == p3)
        assert p1 != p3
        assert hash(p1) == hash(p2)
        assert not(hash(p1) == hash(p3))
        assert not(p1 == 42)
        assert p1 != 42
        p1._pyroRelease()
        p2._pyroRelease()
        p3._pyroRelease()


class TestRemoteMethod:

    class BatchProxyMock(object):
        def __init__(self):
            self.result = []
            self._pyroMaxRetries = 0

        def __copy__(self):
            return self

        def __enter__(self):
            return self

        def __exit__(self, *args):
            pass

        def _pyroClaimOwnership(self):
            pass

        def _pyroInvokeBatch(self, calls, oneway=False):
            self.result = []
            for methodname, args, kwargs in calls:
                if methodname == "error":
                    self.result.append(Pyro5.core._ExceptionWrapper(ValueError("some exception")))
                    break  # stop processing the rest, this is what Pyro should do in case of an error in a batch
                elif methodname == "pause":
                    time.sleep(args[0])
                self.result.append("INVOKED %s args=%s kwargs=%s" % (methodname, args, kwargs))
            if oneway:
                return
            else:
                return self.result

    def testBatchMethod(self):
        proxy = self.BatchProxyMock()
        batch = Pyro5.client.BatchProxy(proxy)
        assert batch.foo(42) is None
        assert batch.bar("abc") is None
        assert batch.baz(42, "abc", arg=999) is None
        assert batch.error() is None  # generate an exception
        assert batch.foo(42) is None  # this call should not be performed after the error
        results = batch()
        result = next(results)
        assert result == "INVOKED foo args=(42,) kwargs={}"
        result = next(results)
        assert result == "INVOKED bar args=('abc',) kwargs={}"
        result = next(results)
        assert result == "INVOKED baz args=(42, 'abc') kwargs={'arg': 999}"
        with pytest.raises(ValueError):
            next(results)
        with pytest.raises(StopIteration):
            next(results)
        assert len(proxy.result) == 4  # should have done 4 calls, not 5
        batch._pyroRelease()

    def testBatchMethodOneway(self):
        proxy = self.BatchProxyMock()
        batch = Pyro5.client.BatchProxy(proxy)
        assert batch.foo(42) is None
        assert batch.bar("abc") is None
        assert batch.baz(42, "abc", arg=999) is None
        assert batch.error() is None  # generate an exception
        assert batch.foo(42) is None  # this call should not be performed after the error
        results = batch(oneway=True)
        assert results is None  # oneway always returns None
        assert len(proxy.result) == 4  # should have done 4 calls, not 5

    def testBatchMethodReuse(self):
        proxy = self.BatchProxyMock()
        batch = Pyro5.client.BatchProxy(proxy)
        batch.foo(1)
        batch.foo(2)
        results = batch()
        assert list(results) == ['INVOKED foo args=(1,) kwargs={}', 'INVOKED foo args=(2,) kwargs={}']
        # re-use the batch proxy:
        batch.foo(3)
        batch.foo(4)
        results = batch()
        assert list(results) == ['INVOKED foo args=(3,) kwargs={}', 'INVOKED foo args=(4,) kwargs={}']
        results = batch()
        assert len(list(results)) == 0



"""
Tests for a running Pyro server, with timeouts.

Pyro - Python Remote Objects.  Copyright by Irmen de Jong (irmen@razorvine.net).
"""

import test_server


class TestServerThreadTimeout(test_server.TestServerThreadNoTimeout):
    SERVERTYPE = "thread"
    COMMTIMEOUT = 2.0

    def testException(self):
        pass


class TestServerMultiplexTimeout(test_server.TestServerMultiplexNoTimeout):
    SERVERTYPE = "multiplex"
    COMMTIMEOUT = 2.0

    def testException(self):
        pass




import time
import threading
from Pyro5.api import expose, behavior, oneway, serve, config


@expose
@behavior(instance_mode="single")
class Server(object):
    def __init__(self):
        self.callcount = 0

    def reset(self):
        self.callcount = 0

    def getcount(self):
        return self.callcount  # the number of completed calls

    def getconfig(self):
        return config.as_dict()

    def delay(self):
        threadname = threading.current_thread().name
        print("delay called in thread %s" % threadname)
        time.sleep(1)
        self.callcount += 1
        return threadname

    @oneway
    def onewaydelay(self):
        threadname = threading.current_thread().name
        print("onewaydelay called in thread %s" % threadname)
        time.sleep(1)
        self.callcount += 1


# main program

config.SERVERTYPE = "undefined"
servertype = input("Servertype threaded or multiplex (t/m)?")
if servertype == "t":
    config.SERVERTYPE = "thread"
else:
    config.SERVERTYPE = "multiplex"


serve({
    Server: "example.servertypes"
})



"""
Name server control tool.

Pyro - Python Remote Objects.  Copyright by Irmen de Jong (irmen@razorvine.net).
"""

from . import errors, core


def handle_command(namesrv, cmd, args):
    def print_list_result(resultdict, title=""):
        print("--------START LIST %s" % title)
        for name, (uri, metadata) in sorted(resultdict.items()):
            print("%s --> %s" % (name, uri))
            if metadata:
                print("    metadata:", metadata)
        print("--------END LIST %s" % title)

    def cmd_ping():
        namesrv.ping()
        print("Name server ping ok.")

    def cmd_listprefix():
        if len(args) == 0:
            print_list_result(namesrv.list(return_metadata=True))
        else:
            print_list_result(namesrv.list(prefix=args[0], return_metadata=True), "- prefix '%s'" % args[0])

    def cmd_listregex():
        if len(args) != 1:
            raise SystemExit("requires one argument: pattern")
        print_list_result(namesrv.list(regex=args[0], return_metadata=True), "- regex '%s'" % args[0])

    def cmd_lookup():
        if len(args) != 1:
            raise SystemExit("requires one argument: name")
        uri, metadata = namesrv.lookup(args[0], return_metadata=True)
        print(uri)
        if metadata:
            print("metadata:", metadata)

    def cmd_register():
        if len(args) != 2:
            raise SystemExit("requires two arguments: name uri")
        namesrv.register(args[0], args[1], safe=True)
        print("Registered %s" % args[0])

    def cmd_remove():
        if len(args) != 1:
            raise SystemExit("requires one argument: name")
        count = namesrv.remove(args[0])
        if count > 0:
            print("Removed %s" % args[0])
        else:
            print("Nothing removed")

    def cmd_removeregex():
        if len(args) != 1:
            raise SystemExit("requires one argument: pattern")
        sure = input("Potentially removing lots of items from the Name server. Are you sure (y/n)?").strip()
        if sure in ('y', 'Y'):
            count = namesrv.remove(regex=args[0])
            print("%d items removed." % count)

    def cmd_setmeta():
        if len(args) < 1:
            raise SystemExit("requires arguments: uri and zero or more meta tags")
        metadata = set(args[1:])
        namesrv.set_metadata(args[0], metadata)
        if metadata:
            print("Metadata updated")
        else:
            print("Metadata cleared")

    def cmd_yplookup_all():
        if len(args) < 1:
            raise SystemExit("requires at least one metadata tag argument")
        print_list_result(namesrv.yplookup(meta_all=args, return_metadata=True), " - searched by metadata")

    def cmd_yplookup_any():
        if len(args) < 1:
            raise SystemExit("requires at least one metadata tag argument")
        print_list_result(namesrv.yplookup(meta_any=args, return_metadata=True), " - searched by metadata")

    commands = {
        "ping": cmd_ping,
        "list": cmd_listprefix,
        "listmatching": cmd_listregex,
        "yplookup_all": cmd_yplookup_all,
        "yplookup_any": cmd_yplookup_any,
        "lookup": cmd_lookup,
        "register": cmd_register,
        "remove": cmd_remove,
        "removematching": cmd_removeregex,
        "setmeta": cmd_setmeta
    }
    try:
        commands[cmd]()
    except Exception as x:
        print("Error: %s - %s" % (type(x).__name__, x))
        raise


def main(args=None):
    from argparse import ArgumentParser
    parser = ArgumentParser(description="Pyro name server control utility.")
    parser.add_argument("-n", "--host", dest="host", help="hostname of the NS", default="")
    parser.add_argument("-p", "--port", dest="port", type=int, help="port of the NS (or bc-port if host isn't specified)")
    parser.add_argument("-u", "--unixsocket", help="Unix domain socket name of the NS")
    parser.add_argument("-v", "--verbose", action="store_true", dest="verbose", help="verbose output")
    parser.add_argument("command", choices=("list", "lookup", "register", "remove", "removematching", "listmatching",
                        "yplookup_all", "yplookup_any", "setmeta", "ping"))
    options, unknown_args = parser.parse_known_args(args)
    if options.verbose:
        print("Locating name server...")
    if options.unixsocket:
        options.host = "./u:" + options.unixsocket
    try:
        namesrv = core.locate_ns(options.host, options.port)
    except errors.PyroError as x:
        print("Error:", x)
        return
    if options.verbose:
        print("Name server found:", namesrv._pyroUri)
    handle_command(namesrv, options.command, unknown_args)
    if options.verbose:
        print("Done.")


if __name__ == "__main__":
    main()



import os
import subprocess
import Pyro5.api


class DmesgServer:
    @Pyro5.api.expose
    def dmesg(self):
        # reading last 20 lines of the kernel's dmesg buffer... (requires root privilege)
        try:
            result = subprocess.check_output(["dmesg", "--nopager", "--level", "info"])
            return result.decode().splitlines()[-20:]
        except subprocess.SubprocessError as x:
            raise OSError("couldn't run the dmesg command in the server: " + str(x))


if __name__ == "__main__":
    print("Server is running as:")
    print(" uid/gid", os.getuid(), os.getgid())
    print(" euid/egid", os.geteuid(), os.getegid())

    if os.getuid() != 0:
        print("Warning: lacking root privileges to run the 'dmesg' command to read the kernel's buffer. "
              "Executing the command will fail. For the desired outcome, run this program as root.")
    else:
        print("Running as root. This is okay as we're just running the 'dmesg' command for you.")

    Pyro5.api.serve({
        DmesgServer: "dmesg"
    }, host="localhost", use_ns=False)



from Pyro5.api import expose, oneway


@expose
class TestClass(object):
    def div(self, arg1, arg2):
        return arg1 / arg2

    def error(self):
        raise ValueError('a valueerror! Great!')

    def error2(self):
        return ValueError('a valueerror! Great!')

    def othererr(self):
        raise RuntimeError('a runtime error!')

    @oneway
    def onewayerr(self):
        raise ValueError('error in oneway call!')

    def complexerror(self):
        x = Foo()
        x.crash()

    def unserializable(self):
        return TestClass.unserializable


class Foo(object):
    def crash(self):
        self.crash2('going down...')

    def crash2(self, arg):
        # this statement will crash on purpose:
        x = arg // 2



import time
from Pyro5.api import expose, locate_ns, Daemon, config


@expose
class TimeoutServer(object):
    def delay(self, amount):
        print("sleeping %d" % amount)
        time.sleep(amount)
        print("done.")
        return "slept %d seconds" % amount


config.COMMTIMEOUT = 0  # the server won't be using timeouts

ns = locate_ns()
daemon = Daemon()
daemon2 = Daemon()
obj = TimeoutServer()
obj2 = TimeoutServer()
uri = daemon.register(obj)
uri2 = daemon2.register(obj2)
ns.register("example.timeout", uri)
ns.register("example.timeout.frozendaemon", uri2)
print("Server ready.")
# Note that we're only starting one of the 2 daemons.
# daemon2 is not started, to simulate connection timeouts.
daemon.requestLoop()
